/**
 * @file  AEChannel.cpp
 * @brief Implementierung der Klasse CAEChannel.
 */
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "AEChannel.h"

#include "../Common/eeprom.h"
#include "../Common/CRC.h"
#include <stdlib.h>
#include <time.h>
//#include <io.h>

#ifdef _WIN32

#define S_ISREG(m) ((m & _S_IFMT) == _S_IFREG)
#define S_ISDIR(m) ((m & _S_IFMT) == _S_IFDIR)

#endif

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CAEChannel::CAEChannel(short channelID) : CAEBuffer(channelID), CDFLog(channelID)
{
	// Kommunikationsbasis von m_channel4 an andere uebergeben
	m_channel1_2.SetChannelBase(m_channel4.GetChannelBase());
	m_channel3.SetChannelBase(m_channel4.GetChannelBase());

	// Datensatzpuffer loeschen
	for (int i=0; i < MAX_MB_PAKET; i++) {
	
		m_pucRecordBuf[i] = 0;
	}

	m_iRecordLength = 0;
	m_iRecordError  = 0;

	// Wiederholungsversuche bei CheckDevice
	m_iDevicePollRetry = 0;

	// Protokollzustaende der angesprochenen Geraete
	memset(m_protocolVersion, 0, sizeof(m_protocolVersion));

	// Bevorzugt AEIII+ pruefen
	m_preferAEIII = 0;
}

CAEChannel::~CAEChannel()
{

}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::OpenAEIII(int iType, char *szComString, int iComValue, int iComTimeOut)
{
	// Bevorzugt AEIII+ pruefen
	m_preferAEIII = 1;

	return Open(IV_DEFAULT_DEVICEID, iType, szComString, iComValue, iComTimeOut);
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::Open(int deviceID, int iType, char *szComString, int iComValue, int iComTimeOut)
{
	int res = IV_ERROR_FAIL;

	// Kanal 4 ist Massgebend
	if (m_channel4.GetType() == CChannelBase::Undefined) {

		if (strcmp(m_channel4.m_commString, szComString)) {

			// Bestimmt die Gueltigkeit der Protokollversion

			// Alle Teilnehmer erhalten unbekanntes Protokoll
			memset(m_protocolVersion, 0, sizeof(m_protocolVersion));
		}

		// Je nach Typ der Verbindung die Klasse initialisieren.
		switch (iType) {

		case 1: // RS232 direkt

			// Schnittstelle oeffnen
			res = m_channel4.Open(CChannelBase::Serial, szComString, iComValue, iComTimeOut);
			break;

		case 3: // TCP/IP

			// Schnittstelle oeffnen
			res = m_channel4.Open(CChannelBase::Socket, szComString, iComValue, iComTimeOut);
			break;

		default: // Unbekannter Schnittstellentyp
			res = IV_ERROR_FAIL;
		}

		// Nach Erfolg die Grundeinstellung wiederherstellen
		if (res == IV_ERROR_SUCCESS) {

			SetDeviceID(deviceID);

			if (strcmp(m_channel4.m_commString, szComString)) {

				// Bestimmt die Gueltigkeit der Protokollversion

				// Alle Teilnehmer erhalten unbekanntes Protokoll
				memset(m_protocolVersion, 0, sizeof(m_protocolVersion));
			}
		}
	} 

	return res;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void CAEChannel::Close(void)
{
	// Kanal 4 ist Massgebend
	m_channel4.Close();

	// Bevorzugt AEIII+ pruefen
	m_preferAEIII = 0;
}

// ##################################################################
/**
 * @brief Ermitteln ob die Schnittstelle geoeffnet ist.
 */
int CAEChannel::IsOpen()
{
	// Channel4 ist Massgebend!
	return m_channel4.IsOpen();
}

// ##################################################################
/**
 * @brief Schnittstelle nach Fehler bereinigen.
 */
void CAEChannel::ResetWithError(int err) 
{
	// Gehaltene Protokollversion auswerten
	switch(m_protocolVersion[(m_channel4.GetDeviceIDRemote() == 0xFE) ? 0 : m_channel4.GetDeviceIDRemote()]) {
	case 0:
	case 1:
	case 2:
	case 3: // Anstehende Daten verwerfen
		if (err == IV_ERROR_CRC_PACKET || err == IV_ERROR_CHANNEL_COMMUNICATION || 
			err == IV_ERROR_RECEIVE) {
		
			m_channel4.Flush();
		}
	case 4: 
		// Protocolversion zuruecksetzen
		if (err == IV_ERROR_CHANNEL_COMMUNICATION || 
			err == IV_ERROR_RECEIVE) {
		
			SetProtocolVersion(GetDeviceIDRemote(), 0);
		}
	}
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::SetDeviceIDRemote(short deviceIDRemote)
{
	short deviceIDPrev = m_channel4.GetDeviceIDRemote();

	m_channel1_2.SetDeviceIDRemote((unsigned char)deviceIDRemote);
	m_channel3.SetDeviceIDRemote(deviceIDRemote);
	m_channel4.SetDeviceIDRemote(deviceIDRemote);

	return deviceIDPrev;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::GetDeviceIDRemote()
{
	return m_channel4.GetDeviceIDRemote();
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::SetDeviceID(short deviceID)
{
	short deviceIDPrev = m_channel4.GetDeviceID();

	m_channel1_2.SetDeviceID((unsigned char)deviceID);
	m_channel3.SetDeviceID(deviceID);
	m_channel4.SetDeviceID(deviceID);

	return deviceIDPrev;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::GetDeviceID()
{
	return m_channel4.GetDeviceID();
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::ComReleaseHandle()
{
	if (m_channel4.GetType() != CChannel4::Serial) {

		return IV_ERROR_FAIL;
	}

	// CChannel4
	m_channel4.ReleaseChannel();

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::ComUseHandle(HANDLE hCom)
{
	if (m_channel4.GetType() != CChannel4::Undefined) {

		return IV_ERROR_FAIL;
	}

	// CChannel4
	m_channel4.AttachChannel(CChannelBase::Serial, hCom, 800);

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
HANDLE CAEChannel::GetComPort()
{
	HANDLE hCom = INVALID_HANDLE_VALUE;

	if (m_channel4.GetType() == CChannelBase::Serial) {

		hCom = m_channel4.m_comm;
	} else if (m_channel4.GetType() == CChannelBase::Socket) {

		hCom = (HANDLE)m_channel4.m_socket;
	}

	return hCom;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::SetComPort(HANDLE hCom, int iType, int iComValue, int iTimeOut)
{
	switch (iType) {
	case 0:
	case 2:
		// Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
		if ((m_channel4.GetType() == CChannelBase::Undefined) || 
			((m_channel4.GetType() == CChannelBase::Serial) && (m_channel4.m_comm == hCom)) ) {

			if (m_channel4.ChangeStateSerial(hCom, iComValue, iTimeOut) != IV_ERROR_SUCCESS) {

				return IV_ERROR_FAIL;
			}

			// CChannel4
			m_channel4.AttachChannel(CChannelBase::Serial, hCom, iTimeOut);

		} else {
		
			return IV_ERROR_FAIL;
		}
		break;

	case 1:
	case 3:
		// Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
		if ((m_channel4.GetType() == CChannelBase::Undefined) || 
			((m_channel4.GetType() == CChannelBase::Socket) && (m_channel4.m_socket == (SOCKET)hCom)) ) {

			// CChannel4
			m_channel4.AttachChannel(CChannelBase::Socket, hCom, iTimeOut);

		} else {
		
			return IV_ERROR_FAIL;
		}
		break;
	}

	return IV_ERROR_SUCCESS;
}

void CAEChannel::SetCBFunction(int (*pFunction)(void))
{
	m_channel1_2.SetCBFunction(pFunction);
	m_channel3.SetCBFunction(pFunction);
	m_channel4.SetCBFunction(pFunction);
}

// Setzen des Logdateinamen.
void CAEChannel::SetLogFileName(const char *szLogFileName)
{
	m_channel4.SetTraceFileName(szLogFileName);
}	


//#########################################################
//####################### HELPER ##########################
//#########################################################

int CAEChannel::Receive(void *buffer, int receiveCount, int *receivedCount)
{
	return m_channel4.ReceiveData(buffer, receiveCount, (int&)*receivedCount);
}

int CAEChannel::Transmit(void *buffer, int transmitCount, int *transmittedCount)
{
	return m_channel4.TransmitData(buffer, transmitCount, (int&)*transmittedCount);
}

// Aufbereiten und senden eines Protokollpakets
int CAEChannel::TransmitPacket(short deviceIDRemote, unsigned char type, short module, short function, unsigned char *buf, int count, int transmitTrials)
{
	return m_channel4.TransmitPacket(deviceIDRemote, type, module, function, buf, count, transmitTrials);
}

// Empfang und aufbereiten eines Protokollpakets
int CAEChannel::ReceivePacket(short deviceID, unsigned char &type, short &module, short &function, unsigned char *buf, int &count, int receiveTrials)
{
	return m_channel4.ReceivePacket(deviceID, type, module, function, buf, count, receiveTrials);
}


// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
unsigned char CAEChannel::GetProtocolVersion(short deviceIDRemote)
{
	switch(m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote]) {
	case 1: // Version 1 des Protokolls
	case 2: // Version 2 des MM-Protokolls
	case 3: // Version 3 des MM-Protokolls
	case 4: // Version 4 des MM-Protokolls
		return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote];

	default: // Ermitteln der Protokollversion
		
		// Bevorzugt auf AEIII+ pruefen?
		if (m_preferAEIII && (m_channel1_2.Request((unsigned char)deviceIDRemote) == TRUE)) { // Pruefen ob Protokollversion 1 vorhanden
			
			return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 1;
			
		} else if (m_channel4.Request(deviceIDRemote) == IV_ERROR_SUCCESS) { // Pruefen ob Protokollversion 4 vorhanden
			
			return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 4;
		} else {
			
			// Hoher Modbustimeout �lterer Firmwarest�nde uebergehen.
			Sleep(100);
			if (m_channel1_2.Request2((unsigned char)deviceIDRemote) == TRUE) { // Pruefen ob Protokollversion 2 vorhanden
				
				return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 2;
			} else {
				// Hoher Modbustimeout �lterer Firmwarest�nde uebergehen.
				Sleep(100);
				if (m_channel1_2.Request((unsigned char)deviceIDRemote) == TRUE) { // Pruefen ob Protokollversion 1 vorhanden
					
					return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 1;
				} else {
					
					// Hoher Modbustimeout �lterer Firmwarest�nde uebergehen.
					//Sleep(100);
					if (m_channel3.Request(deviceIDRemote) == IV_ERROR_SUCCESS) { // Pruefen ob Protokollversion 3 vorhanden
						
						return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 3;
					}
					
					// Hoher Modbustimeout �lterer Firmwarest�nde uebergehen.
					//Sleep(100);
				}
			}
		}
	}

	return 0;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
unsigned char CAEChannel::SetProtocolVersion(short deviceIDRemote, unsigned char version)
{
	unsigned char old = m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote];

	m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = version;

	return old;
}

void CAEChannel::ClearMem()
{
	// Speicher initialisieren
	memset(m_pucEepMem, 0xFF, sizeof(m_pucEepMem));
	memset(m_pucFlashMem, 0xFF, sizeof(m_pucFlashMem));
	memset(m_pucConfigMem, 0x00, sizeof(m_pucConfigMem));
	memset(m_pucExtendedMem, 0x00, sizeof(m_pucExtendedMem));
}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

int CAEChannel::CollectData(short deviceIDRemote, int *piError)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	// Loeschen des Datenpuffer
	if (m_pucRecordBuf[0]) {

		for (int i=0; i < MAX_MB_PAKET; i++) {
		
			m_pucRecordBuf[i] = 0;
		}

		m_iRecordLength = 0;
		m_iRecordError = 0;
	}

	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		// Datensatz ermitteln
		iRes = m_channel1_2.ReadRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		break;

	case 3: {
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)deviceIDRemote;

		iRes = m_channel3.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], (unsigned short&)m_iRecordLength, status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				m_iRecordError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				m_iRecordError = IV_ERROR_PROCESSING;
			}

			// Fehler
			iRes = -1;
		} else {

			// OK
			iRes = status;

			// Groesze anpassen
			m_iRecordLength += 5;
		}
		break;}

	case 4: {
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

		iRes = m_channel4.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				m_iRecordError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				m_iRecordError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = -1;
		} else {

			// OK
			iRes = status;

			// Groesze anpassen
			m_iRecordLength += 5;
		}
		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return -1;

	}

	switch ( iRes ) {
	case 0: // Keine Datensaetze vorhanden
		
		memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
		m_iRecordLength = 0;
		m_iRecordError = 0;

		return 0;

	case 1: // Online Datensatz
	case 2: // Wiederholter Online Datensatz

		return 2;

	case 3: // Offline Datensatz
	case 4: // Wiederholter Offline Datensatz

		return 1;

	default:
		// Fehler bei Ermittlung
		*piError = m_iRecordError;
	}

	return -1;
}

int CAEChannel::GetDatensatz(unsigned char *pucBuf, int *piLength, int *piError)
{
	int iRes;

	// Wenn BusNum des letzten Datensatzpakets 0 ist dann sind 
	// keine gueltigen Daten vorhanden.
	if (m_pucRecordBuf[0] == 0) {

		return 0;
	}

	// Daten in Puffer uebergeben
	unsigned char ucTmp[MAX_MB_PAKET_DATA * 3 + 1];
	int i, iFail;
	// Zu grosze Datensaetze abfangen!
	if ((m_iRecordLength - 6) > MAX_MB_PAKET_DATA) {

		m_iRecordLength -= (m_iRecordLength - 6) - MAX_MB_PAKET_DATA;
	}
	for (i=0; i < (m_iRecordLength - 6); i++) {
	
		pucBuf[i] = m_pucRecordBuf[i + 4];
		//sprintf((char*)&ucTmp[i * 3], "%02X ", (unsigned char)m_pucRecordBuf[i + 4]);
	}
	ucTmp[i * 3] = 0;

	*piLength = m_iRecordLength - 6;

	//theApp.m_log.Append("\t%s\t%i", ucTmp, *piLength);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		// Datensatz quittieren (3x -> Versuchen)
		for (i=0; (i < 3) && (iRes = m_channel1_2.QuitRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf,
			&m_iRecordLength, &m_iRecordError)) == -1; i++);

		// Quittierung erfolgreich?
		if (iRes) {
			
			// Naechsten Datensatz mit Duplikattest anfordern
			for (i=0, iFail=0; (i < 3) && (iRes = m_channel1_2.ReadRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, 
				&m_iRecordLength, &m_iRecordError)); i++);
		} 
		break;

	case 3: {
		unsigned char status = 0;
		// Datensatz quittieren (3x -> Versuchen)
		iRes = m_channel3.SetRecord(GetDeviceIDRemote(), status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				m_iRecordError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				m_iRecordError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = -1;
		} else {

			// OK
			// naechsten Datensatz lesen.
			unsigned char status = 0;
			m_iRecordLength = 0;
			m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

			iRes = m_channel3.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], (unsigned short&)m_iRecordLength, status);
			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				m_iRecordError = iRes;
				iRes = -1;
			} else if (iRes > 0) {

				// Funktionsfehler
				m_iRecordError = IV_ERROR_PROCESSING;
				iRes = -1;
			} else {

				// OK
				iRes = status;

				// Groesze anpassen
				m_iRecordLength += 5;
			}
		}
		break;}

	case 4: {
		unsigned char status = 0;
		// Datensatz quittieren (3x -> Versuchen)
		iRes = m_channel4.SetRecord(GetDeviceIDRemote(), status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				m_iRecordError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				m_iRecordError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = -1;
		} else {

			// OK
			// naechsten Datensatz lesen.
			unsigned char status = 0;
			m_iRecordLength = 0;
			m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

			iRes = m_channel4.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);
			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				m_iRecordError = iRes;
				iRes = -1;
			} else if (iRes > 0) {

				// Funktionsfehler
				m_iRecordError = IV_ERROR_PROCESSING;
				iRes = -1;
			} else {

				// OK
				iRes = status;

				// Groesze anpassen (wegen veralgemeinerter Ausfuehrung am Funktionsanfang)
				m_iRecordLength += 5;
			}
		}
		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return -1;
	}
		
	switch ( iRes ) {
	case 0: // Keine Datensaetzte vorhanden
			// Aktuell uebergebener ist letzter moeglicher, bei Quittierungsmisserfolg.

		memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
		m_iRecordLength = 0;
		m_iRecordError = 0;

		return 2;

		// Online Datensaetze
	case 1: // Online Datensatz
	case 2: // Wiederholter Online Datensatz

		return 1;

		// Offline Datensaetze
	case 3: // Offline Datensatz
	case 4: // Wiederholter Offline Datensatz

		return 1;

	default:
		*piError = m_iRecordError;
	}

	return -1;
}

int CAEChannel::DFReadRecord(int deviceIDRemote, char *buf, int *length, int *error)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
	
	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		// Datensatz ermitteln
		iRes = m_channel1_2.ReadRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		break;

	case 3: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

		iRes = m_channel3.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], (unsigned short&)m_iRecordLength, status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Funktionsfehler
			m_iRecordError = IV_ERROR_PROCESSING;
			iRes = -1;
		} else {

			// OK
			iRes = status;

			// Groesze anpassen
			m_iRecordLength += 5;
		}
		break;}

	case 4: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

		iRes = m_channel4.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Funktionsfehler
			m_iRecordError = IV_ERROR_PROCESSING;
			iRes = -1;
		} else {

			// OK
			iRes = status;

			// Groesze anpassen
			m_iRecordLength += 5;
		}
		break;}

	default:
		*error = IV_ERROR_UNREACHABLE;
		return -1;
	}

//	TRACE("( %i )\n", iRes);
	switch ( iRes ) {
	case 0: // Keine Datensaetze vorhanden
		
		memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
		m_iRecordLength = 0;
		m_iRecordError = 0;

		return 0;

	case 1: // Online Datensatz
	case 2: // Wiederholter Online Datensatz
	case 3: // Offline Datensatz
	case 4: // Wiederholter Offline Datensatz
		
		// Daten in Puffer uebergeben
		int i;
		// Zu gro�e Datensaetze abfangen!
		if ((m_iRecordLength - 6) > MAX_MB_PAKET_DATA) {
			m_iRecordLength -= (m_iRecordLength - 6) - MAX_MB_PAKET_DATA;
		}
		for (i=0; i < (m_iRecordLength - 6); i++) {
		
			buf[i] = m_pucRecordBuf[i + 4];
		}

		*length = m_iRecordLength - 6;

		return iRes;

	default:
		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ResetWithError(*error);
		break;
	}

	return -1;
}

int CAEChannel::DFQuitRecord(int deviceIDRemote, int *error)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
	
	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		// Datensatz ermitteln
		iRes = m_channel1_2.QuitRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		break;

	case 3: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;

		iRes = m_channel3.SetRecord(GetDeviceIDRemote(), status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Funktionsfehler
			m_iRecordError = IV_ERROR_PROCESSING;
			iRes = -1;
		} else {

			// OK
			iRes = status;
		}
		break;}

	case 4: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;

		iRes = m_channel4.SetRecord(GetDeviceIDRemote(), status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			m_iRecordError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Funktionsfehler
			m_iRecordError = IV_ERROR_PROCESSING;
			iRes = -1;
		} else {

			// OK
			iRes = status;
		}
		break;}

	default:
		*error = IV_ERROR_UNREACHABLE;
		return -1;
	}

//	TRACE("( %i )\n", iRes);
	switch ( iRes ) {
	case 0: // Keine Quittung noetig
		
		return 0;

	case 1: // Online Datensatzquit.
	case 2: // Wiederholte Online Datensatzquit.
	case 3: // Offline Datensatzquit.
	case 4: // Wiederholte Offline Datensatzquit.
		
		return iRes;

	default:
		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ResetWithError(*error);
		break;
	}

	return -1;
}

// Loescht alle Datensaetze
int CAEChannel::ClearData(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		return m_channel1_2.ClearData((unsigned char)GetDeviceIDRemote(), piError);
	case 3: // PZEMasterIV mit Protokollversion 3
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;
	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes;
		iRes = m_channel4.ClearRecords(GetDeviceIDRemote());
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::GetDateTime(short deviceIDRemote, unsigned char *pucDateTime, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		if(!m_channel1_2.ReadMem(MB_GET_TIME, 7, pucDateTime, piError)) {
				
			return FALSE;
		}
		return TRUE;

	case 3: 
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: {
		int iRes = IV_ERROR_SUCCESS;
		char szTime[20];
		iRes = m_channel4.GetRTC(GetDeviceIDRemote(), szTime);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;

			pucDateTime[0] = (szTime[0] - 0x30) * 10 + szTime[1] - 0x30;
			pucDateTime[1] = (szTime[2] - 0x30) * 10 + szTime[3] - 0x30;

			pucDateTime[2] = (szTime[5] - 0x30) * 10 + szTime[6] - 0x30;

			pucDateTime[3] = (szTime[8] - 0x30) * 10 + szTime[9] - 0x30;

			pucDateTime[4] = (szTime[11] - 0x30) * 10 + szTime[12] - 0x30;
			pucDateTime[5] = (szTime[14] - 0x30) * 10 + szTime[15] - 0x30;
			pucDateTime[6] = (szTime[17] - 0x30) * 10 + szTime[18] - 0x30;

		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::SetDateTime(short deviceIDRemote, unsigned char *pucDateTime, int *piError)
{
	unsigned char ucDateTime[8];

	if (pucDateTime == NULL ||
		(pucDateTime[0] != 20) ||
		(pucDateTime[1] > 99) ||
		(pucDateTime[2] < 1) || (pucDateTime[2] > 12) || 
		(pucDateTime[3] < 1) || (pucDateTime[3] > 31) ||
		(pucDateTime[4] > 23) || (pucDateTime[5] > 59) || (pucDateTime[6] > 59)) {
	
		time_t tm_time;
		struct tm *ptm;
		tm_time = time(NULL);
		ptm = localtime(&tm_time);
		
		ucDateTime[0] = (ptm->tm_year + 1900) / 100;
		ucDateTime[1] = (ptm->tm_year + 1900) % 100;
		ucDateTime[2] = ptm->tm_mon + 1;
		ucDateTime[3] = ptm->tm_mday;
		ucDateTime[4] = ptm->tm_hour;
		ucDateTime[5] = ptm->tm_min;
		ucDateTime[6] = ptm->tm_sec;

	} else {

		memcpy(ucDateTime, pucDateTime, 7);
	}

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;
	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		if(!m_channel1_2.WriteMem(MB_SET_TIME, 7, ucDateTime, piError)) {
				
			return FALSE;
		}

		return TRUE;

	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: {
		char szTime[20];

		sprintf(szTime, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
			ucDateTime[0], ucDateTime[1], ucDateTime[2], ucDateTime[3], 
			ucDateTime[4], ucDateTime[5], ucDateTime[6]);

		iRes = m_channel4.SetRTC(GetDeviceIDRemote(), szTime);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::SendInfotext(short deviceIDRemote, char *szMessage, int iLength, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: {
		unsigned char Buffer[245];
		int counter;

		Buffer[0] = 0xFF;
		Buffer[1] = 0;
		Buffer[2] = 0;

		for(counter=0; counter < iLength; counter++) {

			Buffer[counter + 3] = szMessage[counter];
		}

		if(!m_channel1_2.WriteMem(INFOTEXT_MEMORY, iLength + 3, &Buffer[0], piError)) {
			
			return FALSE;
		}

		return TRUE;}

	case 2:
	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;

	case 4: {
		int iRes = IV_ERROR_SUCCESS;

		iRes = m_channel4.ShowInfo(GetDeviceIDRemote(), 0, 0, 0, 0, 0, szMessage);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::SendMessage(short deviceIDRemote, unsigned char ucDelayTime, 
							 unsigned char ucAction,  unsigned char ucAudio, 
							 char *szMessage, int iLength, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: {
	case 2:
		unsigned char Buffer[245];
		int counter;

		Buffer[0] = ucDelayTime;
		Buffer[1] = ucAction;
		Buffer[2] = ucAudio;

		for(counter=0; (counter < iLength) && (counter < (sizeof(Buffer) - 3)); counter++) {
			
			Buffer[3 + counter] = szMessage[counter];
		}

		if(!m_channel1_2.WriteMem(DISPLAY_MEMORY, counter + 3, &Buffer[0], piError)) {
			
			return FALSE;
		}
		return TRUE;}

	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: {
		int iRes = IV_ERROR_SUCCESS;

		iRes = m_channel4.ShowMessage(GetDeviceIDRemote(), ucDelayTime, ucAction, ucAudio, szMessage);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// Datensatzbeschreibung oder Listenbeschreibung aus dem Geraet lesen.
int CAEChannel::GetTableDescription(short deviceIDRemote, unsigned char ucList, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:  // AEIII+
	case 2: {// PZE - Master IV
		ULONG pDat, pFeld, pList[MAX_DATENSATZ_BESCHREIBUNG], pListEnd;
		int i, j;

		// EEPROM lesen 
 		if (!m_channel1_2.ReadMem(MB_START_EEP, EEP_SIZE_SU, m_pucEepMem, piError)) {

			return FALSE;
		}

		// Start und Ende der Datensatzbeschreibungen lesen
		ChangeOrderLong(&m_pucEepMem[P_DAT_STRUCT]);
		StartDatensatzB = *(ULONG*)(&m_pucEepMem[P_DAT_STRUCT]);

		ChangeOrderLong(&m_pucEepMem[P_LIST_STRUCT]);
		EndDatensatzB = *(ULONG*)(&m_pucEepMem[P_LIST_STRUCT]);

		if (StartDatensatzB >=  MAX_FLASH_MEM || EndDatensatzB >= MAX_FLASH_MEM) {
			
			*piError = IV_ERROR_CHANNEL_COMMUNICATION;
			return FALSE;
		}

		if (ucList) {

			//Pointer der Listenbeschreibungen lesen (Ende Datensatzbeschreibung ist anfang Listenb.
			if (!m_channel1_2.ReadMem(EndDatensatzB, MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG), &m_pucFlashMem[EndDatensatzB - FLASH_KONFIG_START], piError)) {
			
				return FALSE;
			}

			// Pointer extrahieren und Anzahl Listenb. ermitteln
			for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

				ChangeOrderLong(&m_pucFlashMem[EndDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);
				pList[i] = *(ULONG*)(&m_pucFlashMem[EndDatensatzB - FLASH_KONFIG_START + i *sizeof(ULONG)]);

				if (pList[i] == 0ul) {
					
					MaxListB = i; 
					i = MAX_DATENSATZ_BESCHREIBUNG; 
					continue;
				}
			}

			// Ende Listenbereich angenaehert berechnen (mit Reserve)
			pListEnd = pList[MaxListB - 1] + sizeof(DatensatzBeschreibung[0]) + 
						(sizeof(DatensatzFeld[0][0]) + 
						sizeof(ULONG)) * (MAX_DATENSATZ_BESCHREIBUNG + 1);

			// Rest der Listenbeschreibung aus AEIII+ lesen
			pDat = EndDatensatzB + MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG);

			if (!m_channel1_2.ReadMem(pDat, pListEnd - pDat, &m_pucFlashMem[pDat - FLASH_KONFIG_START], piError)) {
				
				return FALSE;
			}

			// Listenbeschreibungen auswerten
			for (i=0; i < MaxListB; i++) {

				CpyFromKonfigBuf((unsigned char*)&ListenBeschreibung[i], m_pucFlashMem, pList[i], sizeof(ListenBeschreibung[i]));
				for (j=0; j < ListenBeschreibung[i].Felder; j++) { // Feldbeschreibung einlesen
					
					ChangeOrderLong  (&m_pucFlashMem[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);
					pFeld = *(ULONG*)(&m_pucFlashMem[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);

					ASSERT (pFeld + sizeof(DatensatzFeld[i][j]) < pListEnd);

					CpyFromKonfigBuf((unsigned char*)&ListenFeld[i][j], m_pucFlashMem, pFeld, sizeof(ListenFeld[i][j]));
				}
			}
		} else {

			// Datensatzbeschreibungen lesen
			if (!m_channel1_2.ReadMem(StartDatensatzB, EndDatensatzB - StartDatensatzB, &m_pucFlashMem[StartDatensatzB - FLASH_KONFIG_START], piError)) {
				
				return FALSE;
			}

			// Datensatzbeschreibung auswerten
			for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

				ChangeOrderLong(&m_pucFlashMem[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);
				pDat = *(ULONG*)(&m_pucFlashMem[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);

				if (pDat == 0ul) {
					
					MaxDatB = i; 
					i = MAX_DATENSATZ_BESCHREIBUNG; 
					continue;
				}

				CpyFromKonfigBuf((unsigned char*)&DatensatzBeschreibung[i], m_pucFlashMem, pDat, sizeof(DatensatzBeschreibung[i]));
				
				// Feldbeschreibung einlesen
				for (j=0; j < DatensatzBeschreibung[i].Felder; j++) { 
				
					ChangeOrderLong(&m_pucFlashMem[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);
					pFeld = *(ULONG*)(&m_pucFlashMem[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);

					CpyFromKonfigBuf((unsigned char*)&DatensatzFeld[i][j], m_pucFlashMem, pFeld, sizeof(DatensatzFeld[i][j]));
				}
			}
		}		
		break;}

	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: {

		int iRes = GetTables4(GetDeviceIDRemote(), (ucList) ? (char*)"LIST" : (char*)"RECORD");
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return FALSE;
	}

	return TRUE;
}

// Datensatzbeschreibungen aus Geraet lesen
int CAEChannel::GetTables3(short deviceIDRemote, char *typeName)
{
	char name[64];
	unsigned char tableCount, fieldCount, i;
	int err = IV_ERROR_SUCCESS;

	STable2 table;
	STable2Field field;
	TDatensatzBeschreibung *pTable;
	TDatensatzFeld *pField;
	
	// Struktur uebernehmen
	if (strcmp(typeName, "LIST") == 0) {

		MaxListB = 0;
	} else if (strcmp(typeName, "RECORD") == 0) {

		MaxDatB = 0;
	} else {

		return -IV_ERROR_LIST_PROCESSING;
	}

	// Alle Tabellenbeschreibungen lesen
	for (tableCount=0; tableCount < MAX_DATENSATZ_BESCHREIBUNG; tableCount++) {

		sprintf(name, "%s.%i", typeName, tableCount);

		if ((err = m_channel3.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS) {

			// Nur wenn der Minorname (Index) nicht bekannt ist, 
			// liegt wirklich ein Fehler vor.
			if (err < 0) {

				// Kommunikationsfehler
				return err;

			} else if (err != 2) {

				// Funktionsfehler
				return err;
			}

			break;
		}

		// Struktur uebernehmen
		if (strcmp(typeName, "LIST") == 0) {

			pTable = &ListenBeschreibung[tableCount];
		} else if (strcmp(typeName, "RECORD") == 0) {

			pTable = &DatensatzBeschreibung[tableCount];
		}

		strcpy((char*)pTable->Name, table.name);
		pTable->Felder = table.count;
		pTable->P = 0;
	}

	// Feldbeschreibung lesen
	for (i=0; i < tableCount; i++) {

		// Struktur uebernehmen
		if (strcmp(typeName, "LIST") == 0) {

			pTable = &ListenBeschreibung[i];
		} else if (strcmp(typeName, "RECORD") == 0) {

			pTable = &DatensatzBeschreibung[i];
		}

		// Alle Felder lesen
		for (fieldCount=0; fieldCount < pTable->Felder; fieldCount++) {

			sprintf(name, "%s.%i.%i", typeName, i, fieldCount);

			if ((err = m_channel3.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS) {

				if (err < 0) {

					// Kommunikationsfehler
					return err;

				} else {

					// Funktionsfehler
					return err;
				}
			}

			// Struktur uebernehmen
			if (strcmp(typeName, "LIST") == 0) {

				pField = &ListenFeld[i][fieldCount];
			} else if (strcmp(typeName, "RECORD") == 0) {

				pField = &DatensatzFeld[i][fieldCount];
			}

			pField->Typ = field.type;
			pField->Cnt = field.size;
			strcpy((char*)pField->Name, field.name);
		}
	}

	// Information uebernehmen
	if (strcmp(typeName, "LIST") == 0) {

		MaxListB = tableCount;
	} else if (strcmp(typeName, "RECORD") == 0) {

		MaxDatB = tableCount;
	}

	return err;
}

// Datensatzbeschreibungen aus Geraet lesen
int CAEChannel::GetTables4(short deviceIDRemote, char *typeName)
{
	char name[64];
	unsigned char tableCount, fieldCount, i;
	int err = IV_ERROR_SUCCESS;

	STable2 table;
	STable2Field field;
	TDatensatzBeschreibung *pTable;
	TDatensatzFeld *pField;
	
	// Struktur uebernehmen
	if (strcmp(typeName, "LIST") == 0) {

		MaxListB = 0;
	} else if (strcmp(typeName, "RECORD") == 0) {

		MaxDatB = 0;
	} else {

		return -IV_ERROR_LIST_PROCESSING;
	}

	// Alle Tabellenbeschreibungen lesen
	for (tableCount=0; tableCount < MAX_DATENSATZ_BESCHREIBUNG; tableCount++) {

		sprintf(name, "%s.%i", typeName, tableCount);

		if ((err = m_channel4.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS) {

			// Nur wenn der Minorname (Index) nicht bekannt ist, 
			// liegt wirklich ein Fehler vor.
			if (err < 0) {

				// Kommunikationsfehler
				return err;

			} else if (err == 2) { // Tabellenname unbekannt

				// Fehlercode zuruecksetzen
				err = IV_ERROR_SUCCESS;

			} else {

				// Funktionsfehler
				return err;
			}

			break;
		}

		// Struktur uebernehmen
		if (strcmp(typeName, "LIST") == 0) {

			pTable = &ListenBeschreibung[tableCount];
		} else if (strcmp(typeName, "RECORD") == 0) {

			pTable = &DatensatzBeschreibung[tableCount];
		}

		strcpy((char*)pTable->Name, table.name);
		pTable->Felder = table.count;
		pTable->P = 0;
	}

	// Feldbeschreibung lesen
	for (i=0; i < tableCount; i++) {

		// Struktur uebernehmen
		if (strcmp(typeName, "LIST") == 0) {

			pTable = &ListenBeschreibung[i];
		} else if (strcmp(typeName, "RECORD") == 0) {

			pTable = &DatensatzBeschreibung[i];
		}

		// Alle Felder lesen
		for (fieldCount=0; fieldCount < pTable->Felder; fieldCount++) {

			sprintf(name, "%s.%i.%i", typeName, i, fieldCount);

			if ((err = m_channel4.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS) {

				if (err < 0) {

					// Kommunikationsfehler
					return err;

				} else {

					// Funktionsfehler
					return err;
				}
			}

			// Struktur uebernehmen
			if (strcmp(typeName, "LIST") == 0) {

				pField = &ListenFeld[i][fieldCount];
			} else if (strcmp(typeName, "RECORD") == 0) {

				pField = &DatensatzFeld[i][fieldCount];
			}

			pField->Typ = field.type;
			pField->Cnt = field.size;
			strcpy((char*)pField->Name, field.name);
		}
	}

	// Information uebernehmen
	if (strcmp(typeName, "LIST") == 0) {

		MaxListB = tableCount;
	} else if (strcmp(typeName, "RECORD") == 0) {

		MaxDatB = tableCount;
	}

	return err;
}

// Listenbeschreibung aus dem Geraet lesen.
int CAEChannel::GetListName(short deviceIDRemote, unsigned char ucList, char *name, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	ULONG add;
	TDatensatzBeschreibung dsb;

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		break;

	case 3:
	case 4:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return FALSE;
	}

	if (!m_channel1_2.SetDeviceIDRemote((unsigned char)deviceIDRemote)) {

		*piError = IV_ERROR_DEVICE_ID_REMOTE;
		return FALSE;
	}

	// Die Addresse auf die Listenbeschreibungen aus dem EEPROM-Bereich lesen 
 	if (!m_channel1_2.ReadMem(MB_START_EEP + P_LIST_STRUCT, sizeof(ULONG), (UCHAR*)&add, piError)) {

		return FALSE;
	}
	ChangeOrderLong((UCHAR*)&add);

	// Die Addresse auf die Listenbeschreibung lesen.
	if (!m_channel1_2.ReadMem(add + ucList * sizeof(ULONG), sizeof(ULONG), (UCHAR*)&add, piError)) {
	
		return FALSE;
	}
	ChangeOrderLong((UCHAR*)&add);


	// Wenn die Rueckgabe FALSE und Error = IV_ERROR_SUCCESS, dann war die Liste nicht vorhanden.
	*piError = IV_ERROR_SUCCESS;

	if (add) {

		// Die Listenbeschreibung lesen.
		if (!m_channel1_2.ReadMem(add, sizeof(TDatensatzBeschreibung), (UCHAR*)&dsb, piError)) {
		
			return FALSE;
		}

		strcpy(name, (char*)dsb.Name);
		return TRUE;
	}

	return FALSE;
}

int CAEChannel::UpLoadSetupData(short deviceIDRemote, char *szFileName, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int i;
//	char path_buffer[_MAX_PATH]; // 260
	char drive[_MAX_DRIVE];      // 3
	char dir[_MAX_DIR];          // 256
	char fname[_MAX_FNAME];      // 256
	char ext[_MAX_EXT];          // 256

	// Dateiendung pruefen
#ifdef _WIN32
	_splitpath(szFileName, drive, dir, fname, ext);
#else
	strcpy(ext, &szFileName[strlen(szFileName) - 4]);
#endif
	for (i=0; ext[i] && i < 4; i++) ext[i] = tolower(ext[i]);
	if (strcmp(ext, ".aes") != 0) {

		// Die Anwendung kann �ber den gew�hlten Weg, nur Dateien vom Typ (*.aes) �ffnen.
		*piError = IV_ERROR_FILE_EXTENSION;
		return FALSE;
	}

	// Setup-Datei oeffnen
	FILE *pFile = fopen(szFileName, "rb");

	if (pFile == NULL) {

		*piError = IV_ERROR_FILE_OPEN;
		return FALSE;
	}
	
	// Erweiterter Speicherbereich
	struct stat statBuf;
	int res = stat(szFileName, &statBuf);

	// Dateigroesse pruefen
	if ((res != 0) || (statBuf.st_size < 
		(EEP_SIZE + FLASH_SIZE_SU + 20 + CONFIG_SIZE_SU + EXTENDED_SIZE_SU))) {

		fclose(pFile);

		*piError = IV_ERROR_FILE_FORMAT;
		return FALSE;
	}

	// Speicher initialisieren
	ClearMem();

	// EEPROM - lesen
	fread(m_pucEepMem, sizeof(unsigned char), EEP_SIZE, pFile);

	// Flashbereich - lesen
	fread(m_pucFlashMem, sizeof(unsigned char), FLASH_SIZE_SU, pFile);
	
	// ListFile - lesen
	int iLen;
	unsigned char buf[0xFFF];
	for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {
	
		fread(buf, sizeof(unsigned char), 1, pFile); // ein byte laenge
		if (buf[0] == 0xFF) {

			fread(buf, sizeof(unsigned char), 2, pFile); // zwei byte laenge
			iLen = *(unsigned short*)buf;

		} else {

			iLen = buf[0];
		}

		fread(buf, sizeof(unsigned char), iLen, pFile);
	}

	// Konfigurationsdaten - lesen
	fread(m_pucConfigMem, sizeof(unsigned char), CONFIG_SIZE_SU, pFile);

	// Erweiterter Flashbereich - lesen
	fread(m_pucExtendedMem, sizeof(unsigned char), EXTENDED_SIZE_SU, pFile);

	// Pruefen der Extended - Struktur
	TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

	if (extBegin.ulCheck != EXTENDED_CHECK) { // existiert?

		fclose(pFile);

		*piError = IV_ERROR_FILE_FORMAT;
		return FALSE;
	}

	fclose(pFile);

	// Pruefen ob Setupdatei fuer diesen Geraetetyp ist.
	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: // AEIII+
		if (m_pucEepMem[P_DEVICE_TYPE] != 0) {

			*piError = IV_ERROR_DEVICE_WRONG;
			return FALSE;
		}
		break;

	case 2: // PZE Master IV
		if (m_pucEepMem[P_DEVICE_TYPE] != 1) {

			*piError = IV_ERROR_DEVICE_WRONG;
			return FALSE;
		}
		break;

	case 3: // PZE Master IV
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;

	case 4: { // PZE / ZK / BDE Master IV

		int iRes;
		unsigned char type = 0;

		iRes = GetTypeDevice(GetDeviceIDRemote(), type);

		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			return FALSE;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				// Geraet als PZE Master IV annehmen
				type = 0;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
				return FALSE;
			}
		}

		if ((m_pucEepMem[P_DEVICE_TYPE] - 1) != type) {

			*piError = IV_ERROR_DEVICE_WRONG;
			return FALSE;
		}
		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return FALSE;
	}

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: 
	case 2: {
		unsigned char ucIsData = 0, temp = 1; // fuer Bedienstopp

		// Abfrage ob Daten vorhanden. 
		// (wegen alter FW, nicht mit CheckData ausgefuehrt.)
		if (!m_channel1_2.ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {
			
			return FALSE;
		}
		// Abbruch wenn Daten im Geraet vorliegen
		if (ucIsData) {

			*piError = IV_ERROR_DATA_AVAILABLE;
			return FALSE;
		}

		// Setupgroesse ermitteln
		ULONG ulSetupSize = GetFlashSetupSize();

		// Vorhandene zur moeglichen Groesse pruefen
		if (ulSetupSize > FLASH_SIZE_SU) {

			*piError = IV_ERROR_FILE_OPEN;
			return FALSE;
		}

		// Bedienstopp setzen
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}

		// Groesse der Flashdaten in EepMem eintragen
		*(ULONG*)&m_pucEepMem[P_FLASH_SIZE_SU] = ulSetupSize;
		ChangeOrderLong(&m_pucEepMem[P_FLASH_SIZE_SU]);

		// Groesse der MDE - Eingabeketten in EepMem loeschen.
		*(ULONG*)&m_pucEepMem[P_MDE_EK_SIZE] = 0xFFFFFFFF;
		ChangeOrderLong(&m_pucEepMem[P_MDE_EK_SIZE]);
		
		// Groesse der Zutrittskontroll - Eingabeketten in EepMem loeschen.
		*(ULONG*)&m_pucEepMem[P_ENTRANCE_EK_SIZE] = 0xFFFFFFFF;
		ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);


		// Groesse der MDE - Eingabeketten in EepMem eintragen.
		*(ULONG*)&m_pucEepMem[P_MDE_EK_SIZE] = GetMDEEKFlashSize();
		ChangeOrderLong(&m_pucEepMem[P_MDE_EK_SIZE]);

		ULONG commVersion = GetProtocolVersion2(deviceIDRemote, TRUE, piError);
		if (commVersion == 0) return FALSE;
		if (commVersion != 0x0401) {

			// Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
			*(ULONG*)&m_pucEepMem[P_ENTRANCE_EK_SIZE] = GetEntranceEKFlashSize();
			ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);
		}

	#define MB_DATABLOCKSIZE 32

		// EEPROM - schreiben 
 		if (!m_channel1_2.WriteMem(MB_START_EEP, EEP_SIZE_SU, m_pucEepMem, piError, MB_DATABLOCKSIZE)) {

			return FALSE;
		}

		// Flash - schreiben
		if (!m_channel1_2.WriteMem(FLASH_KONFIG_START, ulSetupSize, m_pucFlashMem, piError, MB_DATABLOCKSIZE)) {

			return FALSE;
		}

		// Konfigurationsdaten - schreiben
		if (!m_channel1_2.WriteMem(MB_START_CONFIG + 2, CONFIG_SIZE_SU - 2, &m_pucConfigMem[2], piError, MB_DATABLOCKSIZE)) {

			return FALSE;
		}

		// IOModule und MDE ab Version 3.0
		if (commVersion != 0x0401) {

			int i;
			TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

			// IOModul - Konfiguration - schreiben
			if (!m_channel1_2.WriteMem(MB_START_IOMODULE, sizeof(TIOModule), &m_pucExtendedMem[extBegin.pulIOModule], piError, MB_DATABLOCKSIZE)) {

				return FALSE;
			}

			// IOModul - DigitalMDE - schreiben
			for (i=0; i < 8; i++) {
				if (!m_channel1_2.WriteMem(MB_DIGITAL_IO + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
					&m_pucExtendedMem[extBegin.pulIODigital + i * sizeof(TDigitalInput)], piError, MB_DATABLOCKSIZE)) {

					return FALSE;
				}
			}

			// Interne MDE - DigitalMDE - schreiben
			for (i=0; i < 4; i++) {
				if (!m_channel1_2.WriteMem(MB_DIGITAL_MDE + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
					&m_pucExtendedMem[extBegin.pulMDEDigital + i * sizeof(TDigitalInput)], piError, MB_DATABLOCKSIZE)) {

					return FALSE;
				}
			}

			// MDE - Eingabeketten - schreiben
			// Bei 4.0 kommt erst 1Byte Anzahl im Array
			// Ab 4.1 kommen erst 4Byte Anzahl im Array
			ULONG ulSize = 0;
			if (commVersion == 0x0400) {

				ulSize = GetMDEEKFlashSize_0400();
			} else {
				
				ulSize = GetMDEEKFlashSize();
			}

			if (!m_channel1_2.WriteMem(MB_MDE_EK, ulSize, &m_pucExtendedMem[extBegin.pulEKArray], piError, MB_DATABLOCKSIZE)) {

				return FALSE;
			}

			// Zutrittskontrolle
			if (commVersion != 0x0400) {

				// Zutrittskontroll - Eingabeketten - schreiben
				if (!m_channel1_2.WriteMem(MB_ENTRANCE_EK, GetEntranceEKFlashSize(), 
					&m_pucExtendedMem[extBegin.pulEKArrayEntrance], piError, MB_DATABLOCKSIZE)) {

					return FALSE;
				}
			}
		}


		temp = 0; // Bedienstopp ruecksetzen
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}

		// Aktuelle Setup - Daten wurden erst durch zuruecksetzen von Bedienstopp uebernommen!

		// Standardwerte fuer Globale Variable
		if (commVersion != 0x0400) {

			int i;
			TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

			// GV setzen?
			if (extBegin.pulValueGVx) {
			
				UCHAR ucNum;
				// Struktur lesen
				TValueGVx ValueGVx = *(TValueGVx*)&m_pucExtendedMem[extBegin.pulValueGVx];

				// Zeit fuer Umschaltung aus Bedienstopp
				Sleep(200);

				for (i=0; i < 8; i++) {

					if (ValueGVx.ucFlags & (1 << i)) {
					
						ucNum = 1 + i;
						if (!SetGlobVar(GetDeviceIDRemote(), &ucNum, 0, ValueGVx.szValue[i], piError)) {
						
							return FALSE;
						}
					}
				}
			}	
		}

		// Standardwerte fuer Systemvariable
		// Ab 04.00.23
		if ((commVersion == 0x0502) ) {

			//############################
			// 2005-09-19: Erweiterte Struktur fuer Parameter
			if (((m_pucEepMem[P_SETUP_VERSION    ] == 3) &&
				 (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
				 ((m_pucEepMem[P_SETUP_VERSION + 2] > 34) ||
				  (m_pucEepMem[P_SETUP_VERSION + 2] == 34) &&
				  (m_pucEepMem[P_SETUP_VERSION + 3] >  5)) ) ||

				 ((m_pucEepMem[P_SETUP_VERSION   ] == 4) &&
				  (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
				  ((m_pucEepMem[P_SETUP_VERSION + 2] > 0) || 
				   (m_pucEepMem[P_SETUP_VERSION + 2] == 0) &&
				   (m_pucEepMem[P_SETUP_VERSION + 3] > 2))) ) {

				TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

				// Liegt die Struktur vor?
				if (extBegin.pulExtended) {

					TExtended extParam = *(TExtended*)&m_pucExtendedMem[extBegin.pulExtended];

					unsigned char value[32], name[64];
					// Buchungswiederholsperrdauer
					sprintf((char*)name, "_SERVICE.DBLINPUT");
					sprintf((char*)value, "%i", extParam.doubleInputTime);
					// 
					if (!SetGlobVar(GetDeviceIDRemote(), name, 2, (char*)value, piError)) {
					
						if (// Fehler liegt bei Wert
							*piError != IV_ERROR_VARIABLE_VALUE &&
							// Variable nicht bekannt
							*piError != IV_ERROR_VARIABLE_UNDEFINED)

						return FALSE;
					}
					// Abblendedauer fuer Bestaetigungsfeld
					sprintf((char*)name, "_SERVICE.DEMANDTIMER");
					sprintf((char*)value, "%i", extParam.confirmationTime);
					// 
					if (!SetGlobVar(GetDeviceIDRemote(), name, 2, (char*)value, piError)) {
					
						if (// Fehler liegt bei Wert
							*piError != IV_ERROR_VARIABLE_VALUE &&
							// Variable nicht bekannt
							*piError != IV_ERROR_VARIABLE_UNDEFINED)

						return FALSE;
					}
				}
			}
		}

		return TRUE;}

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes = IV_ERROR_SUCCESS;

		if ((iRes = CheckData(GetDeviceIDRemote(), piError)) == -1) {

			return FALSE;
		} else if (iRes) {

			*piError = IV_ERROR_DATA_AVAILABLE;
			return FALSE;
		}

		TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;
		//############################
		// 2007-07-06: Struktur geaendert wegen weiterer benoetigter Werte.
		if ((m_pucExtendedMem[P_SETUP_VERSION] == 0xFF) ||
			((m_pucExtendedMem[P_SETUP_VERSION    ] == 4) &&
			(m_pucExtendedMem[P_SETUP_VERSION + 1] == 1) &&
			(m_pucExtendedMem[P_SETUP_VERSION + 2] == 4) && 
			(m_pucExtendedMem[P_SETUP_VERSION + 3] == 0)) ) {
			
			// evntl. vorhandene Strukturen loeschen
			extBegin.pulMDEAnalog = 0;
		} 

		iRes = UploadSetup4(GetDeviceIDRemote());
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}

			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::UploadSetup4(short deviceIDRemote)
{
	int i, res = IV_ERROR_SUCCESS;

	//@@: Uebergangsweise zur 04.01.01 release
	unsigned char buf[0x400];
	unsigned char countTable, countField;
	unsigned long sizeAccess;

	//////////////////////////////
	// Vorbereitung (Berechnung von Groessen und CRC)
	unsigned long sizeSetup = GetFlashSetupSize();

	// Vorhandene zur moeglichen Groesse pruefen
	if (sizeSetup > FLASH_SIZE_SU) {

		return -IV_ERROR_FILE_FORMAT;
	}

	unsigned long sizeEK = GetEntrance2EKFlashSize();
	// Zeigeroffsets fuer Firmware anpassen
	if (sizeEK > sizeof(buf)) {
		// Puffer zu klein, also kein Zutritt
		return -IV_ERROR_FILE_FORMAT;
	}
	memset(buf, 0, sizeof(buf));
	

	//@@: Uebergangsweise zur 04.01.01 release
	// Zeiger auf Eingabeketten ohne FLASH_KONFIG_START
	// und die Zeiger der Eingabeketten mit einem Offset versehen.
	PrepareEntrance2EK(sizeSetup, buf);

	// Groesse der Flashdaten in EepMem eintragen
	*(unsigned long*)&m_pucEepMem[P_FLASH_SIZE_SU] = sizeSetup;
	ChangeOrderLong(&m_pucEepMem[P_FLASH_SIZE_SU]);

	// Anfangsadresse der Zutrittseingabeketten eintragen
	*(unsigned long*)&m_pucEepMem[P_IC_ACCESS] = sizeSetup;
	ChangeOrderLong(&m_pucEepMem[P_IC_ACCESS]);

	// Anfangsadresse der Zutrittskontrolllisten eintragen
	*(unsigned long*)&m_pucEepMem[P_LIST_ACCESS] = sizeSetup + sizeEK;
	ChangeOrderLong(&m_pucEepMem[P_LIST_ACCESS]);

	// Groesse der MDE - Eingabeketten in EepMem eintragen.
	*(unsigned long*)&m_pucEepMem[P_MDE_EK_SIZE] = GetMDEEKFlashSize();
	//ChangeOrderLong(&m_pucEepMem[P_MDE_EK_SIZE]);
	
	// Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
	*(unsigned long*)&m_pucEepMem[P_ENTRANCE_EK_SIZE] = sizeEK;
	ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);


	//
	unsigned long address, size, crc;
	address = size = crc = 0;

	// Segmentbegin: Bedienung stoppen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	
	///////////////
	// Grunddaten

	// CRC berechnen
	crc = 0xFFFFFFFF;
	crc = m_channel4.GetCRC32(m_pucEepMem, EEP_SIZE_SU, crc, FALSE);
	crc = m_channel4.GetCRC32(m_pucConfigMem, 0x80, crc, FALSE);
	crc = m_channel4.GetCRC32(m_pucFlashMem, sizeSetup, crc, FALSE);
	//@@: Uebergangsweise zur 04.01.01 release
	crc = m_channel4.GetCRC32(buf, sizeEK, crc, FALSE);

	
	///////////////
	// ZK II

	// Anzahl der Zutrittslisten ermitteln
	for (i=0, countTable=0; countTable < (sizeof(m_entrance2Pointer) / sizeof(unsigned long)); countTable++) {

		if (m_entrance2Pointer[countTable] == 0) {

			break;
		}

		// Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
		i += m_entrance2Table[countTable].count;
	}
	sizeAccess = sizeof(m_entrance2Pointer);
	sizeAccess += countTable * sizeof(STable2);
	sizeAccess += i * sizeof(STable2Field);

	// CRC der Zutrittslisten ermitteln
	crc = m_channel4.GetCRC32((unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer), crc, FALSE);

	for (i=0; i < countTable; i++) {

		crc = m_channel4.GetCRC32((unsigned char*)&m_entrance2Table[i], sizeof(STable2), crc, FALSE);

		for (countField=0; countField < m_entrance2Table[i].count; countField++) {

			// Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
			crc = m_channel4.GetCRC32((unsigned char*)&m_entrance2TableField[i][countField], sizeof(STable2Field), crc, ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))));
		}
	}

	//////////////////////////////
	// Begin

	// Gesamtgroesze berechnen
	size = EEP_SIZE_SU + 0x80 + sizeSetup + sizeEK + sizeAccess;

	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}


	//////////////////////////////
	// Grunddaten

	// EEPROM uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucEepMem, EEP_SIZE_SU)) != IV_ERROR_SUCCESS) {

		return res;
	}
	// CONFIG uebertragen (128 Byte)
	address += EEP_SIZE_SU;
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucConfigMem, 0x80)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// FLASH uebertragen
	address += 0x80;
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucFlashMem, sizeSetup)) != IV_ERROR_SUCCESS) {

		return res;
	}

	//@@: Uebergangsweise zur 04.01.01 release
	//////////////////////////////
	// Pointerarray + Strukturen der Zutrittslisten uebertragen
	
	// Eingabekette der Zutrittskontrolle
	address += sizeSetup;
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, sizeEK)) != IV_ERROR_SUCCESS) {

		return res;
	}

	address += sizeEK;
	// Pointerarray uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer))) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += sizeof(m_entrance2Pointer);

	// Strukturen uebertragen
	for (i=0; i < countTable; i++) {

		// Tabellenstruktur uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&m_entrance2Table[i], sizeof(STable2))) != IV_ERROR_SUCCESS) {

			return res;
		}
		address += sizeof(STable2);

		// Feldstrukturen uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2TableField[i], m_entrance2Table[i].count * sizeof(STable2Field))) != IV_ERROR_SUCCESS) {

			return res;
		}
		address += m_entrance2Table[i].count * sizeof(STable2Field);
	}

	//////////////////////////////
	// Ende

	// Segmentend: Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}


	// Extendedstruktur lesen
	TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

	///////////////
	// Extended - Speicher Vorbereitung

	// Zeigeroffsets anpassen (substrahieren von Offset)
	AdjustExtendedMemPointer();
	
	// Groesse MDE und dessen Eingabeketten
	size = extBegin.pulFreeScope;

	// CRC berechnen
	crc = 0xFFFFFFFF;
	// IOModul - Konfiguration
	crc = m_channel4.GetCRC32(m_pucExtendedMem, size, crc, TRUE);

	//////////////////////////////
	// Begin

	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS) {

		if ((res & 0xFF) == 1) {
			// Funktionsfehler Typ IV_VALUE_SEGMENT_SETUP_EXTENDED unbekannt.
		} else {
		
			return res;
		}
	}

	if (res == IV_ERROR_SUCCESS) {

		//////////////////////////////
		// Extended - Speicher Uebertragen

		if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucExtendedMem, size)) != IV_ERROR_SUCCESS) {

			return res;
		}
		address += size;

		//////////////////////////////
		// Ende

		// Segmentend: Setup uebertragen
		if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}
	}

	// Zeigeroffsets anpassen (addieren von Offset)
	AdjustExtendedMemPointer();

	crc = 0;
	// Segmentend: Bedienung stoppen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// GV setzen?
	if (extBegin.pulValueGVx) {
	
		SGlobVar *globVar = (SGlobVar*)m_pucFlashMem;
		char name[64];
		memset(name, 0, sizeof(name));

		// Struktur lesen
		TValueGVx ValueGVx = *(TValueGVx*)&m_pucExtendedMem[extBegin.pulValueGVx];

		for (i=0; i < 8; i++) {

			if (ValueGVx.ucFlags & (1 << i)) {
			
				sprintf(name, "SETUP.%s", &globVar->SFeld[i]);
				if ((res = m_channel4.SetVariable(deviceIDRemote, name, (char*)&ValueGVx.szValue[i])) != IV_ERROR_SUCCESS) {
				
					// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
					// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
					if ((res & 0xFF) == 3) {
						// Fehler liegt bei Wert
						//res = -IV_ERROR_VARIABLE_VALUE;
					} else if ((res & 0xFF) == 2) {
						// Variable nicht bekannt
						//res = -IV_ERROR_VARIABLE_UNDEFINED;
					}

					return res;
				}
			}
		}
	}	

	// Standardwerte fuer Systemvariable
	//############################
	// 2005-09-19: Erweiterte Struktur fuer Parameter
	if (((m_pucEepMem[P_SETUP_VERSION    ] == 3) &&
		 (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
		 ((m_pucEepMem[P_SETUP_VERSION + 2] > 34) ||
		  (m_pucEepMem[P_SETUP_VERSION + 2] == 34) &&
		  (m_pucEepMem[P_SETUP_VERSION + 3] >  5)) ) ||

		 ((m_pucEepMem[P_SETUP_VERSION   ] == 4) &&
		  (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
		  ((m_pucEepMem[P_SETUP_VERSION + 2] > 0) || 
		   (m_pucEepMem[P_SETUP_VERSION + 2] == 0) &&
		   (m_pucEepMem[P_SETUP_VERSION + 3] > 2))) ) {

		TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

		// Liegt die Struktur vor?
		if (extBegin.pulExtended) {

			TExtended extParam = *(TExtended*)&m_pucExtendedMem[extBegin.pulExtended];

			char value[32], name[64];
			// Buchungswiederholsperrdauer
			sprintf(name, "_SERVICE.DBLINPUT");
			if (extParam.doubleInputTime > 65499) {
				extParam.doubleInputTime = 65499;
			}
			sprintf(value, "%i", extParam.doubleInputTime);
			// 
			if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, value)) != IV_ERROR_SUCCESS) {
			
				// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
				// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
				if ((res & 0xFF) == 3) {
					// Fehler liegt bei Wert
					//res = -IV_ERROR_VARIABLE_VALUE;
					res = IV_ERROR_SUCCESS;
				} else if ((res & 0xFF) == 2) {
					// Variable nicht bekannt
					//res = -IV_ERROR_VARIABLE_UNDEFINED;
					res = IV_ERROR_SUCCESS;
				}

				return res;
			}
			// Abblendedauer fuer Bestaetigungsfeld
			sprintf(name, "_SERVICE.DEMANDTIMER");
			if (extParam.confirmationTime > 255) {
				extParam.confirmationTime = 255;
			}
			sprintf(value, "%i", extParam.confirmationTime);
			// 
			if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, value)) != IV_ERROR_SUCCESS) {
			
				// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
				// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
				if ((res & 0xFF) == 3) {
					// Fehler liegt bei Wert
					//res = -IV_ERROR_VARIABLE_VALUE;
					res = IV_ERROR_SUCCESS;
				} else if ((res & 0xFF) == 2) {
					// Variable nicht bekannt
					//res = -IV_ERROR_VARIABLE_UNDEFINED;
					res = IV_ERROR_SUCCESS;
				}

				return res;
			}
		}
	}

	return res;
}

int CAEChannel::DownLoadSetupData(short deviceIDRemote, char *szFileName, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);


	int res = FALSE;
	// Setup-Datei oeffnen
	FILE *pFile = fopen(szFileName, "wb");

	if (pFile == NULL) {

		*piError = IV_ERROR_FILE_OPEN;
		return FALSE;
	}

	// Speicher initialisieren
	SetDocumentDefault();

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: 
	case 2: {
		unsigned char temp = 1; // fuer Bedienstopp
		unsigned long ulTmp, ulSize = FLASH_SIZE_SU;

		// Bedienstopp setzen.
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}

		// EEPROM - lesen 
 		if (!m_channel1_2.ReadMem(MB_START_EEP, EEP_SIZE_SU, m_pucEepMem, piError)) {

			return FALSE;
		}

		// Groesse der Setupdaten bereits im Eep?
		ulTmp = *(ULONG*)&m_pucEepMem[P_FLASH_SIZE_SU];
		ChangeOrderLong((unsigned char*)&ulTmp);
		// Liegt im moeglichen Bereich?
		if (!(ulTmp > FLASH_SIZE_SU) && !(ulTmp == 0)) {

			ulSize = ulTmp;
		}

		// Flashbereich - lesen
		if (!m_channel1_2.ReadMem(FLASH_KONFIG_START, ulSize, m_pucFlashMem, piError)) {

			return FALSE;
		}

		// Konfiguration - lesen
		if (!m_channel1_2.ReadMem(MB_START_CONFIG, CONFIG_SIZE_SU, m_pucConfigMem, piError)) {

			return FALSE;
		}

		int i;
		TExtendedFlash extBegin;
		// gesetze Struktur uebernehmen
		extBegin = *(TExtendedFlash*)m_pucExtendedMem;

		// IOModul - lesen
		extBegin.pulIOModule = extBegin.pulFreeScope;

		// IOModul - Konfiguration - lesen
		if (!m_channel1_2.ReadMem(MB_START_IOMODULE, sizeof(TIOModule), &m_pucExtendedMem[extBegin.pulIOModule], piError)) {

			return FALSE;
		}

		// Auf naechsten freien Speicher
		extBegin.pulFreeScope += sizeof(TIOModule);

		// Strukturen fuer Digitale Eingaenge des IOModules eintragen.
		extBegin.pulIODigital = extBegin.pulFreeScope;

		// IOModul - DigitalMDE - lesen
		for (i=0; i < 8; i++) {
			if (!m_channel1_2.ReadMem(MB_DIGITAL_IO + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
				&m_pucExtendedMem[extBegin.pulIODigital + i * sizeof(TDigitalInput)], piError)) {

				return FALSE;
			}
		}

		// Auf naechsten freien Speicher
		extBegin.pulFreeScope += 8 * sizeof(TDigitalInput);

		// Strukturen fuer Digitale Eingaenge der MDE eintragen.
		extBegin.pulMDEDigital = extBegin.pulFreeScope;

		// Interne MDE - DigitalMDE - lesen
		for (i=0; i < 4; i++) {
			if (!m_channel1_2.ReadMem(MB_DIGITAL_MDE + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
				&m_pucExtendedMem[extBegin.pulMDEDigital + i * sizeof(TDigitalInput)], piError)) {

				return FALSE;
			}
		}

		// Auf naechsten freien Speicher
		extBegin.pulFreeScope += 4 * sizeof(TDigitalInput);


		// Strukturen der Eingabeketten von MDE eintragen
		extBegin.pulEKArray = extBegin.pulFreeScope;

		// Groesse der MDE - Eingabeketten bereits im Eep?
		ulTmp = *(ULONG*)&m_pucEepMem[P_MDE_EK_SIZE];
		ChangeOrderLong((unsigned char*)&ulTmp);
		// Liegt im moeglichen Bereich?
		if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0)) {

			ulSize = ulTmp;
		} else {
		
			ulSize = EXTENDED_SIZE_SU - extBegin.pulEKArray;
		}


		// MDE - Eingabeketten - lesen
		// Bei 4.0 kommt erst 1Byte Anzahl im Array
		// Ab 4.1 kommen erst 4Byte Anzahl im Array
		if (!m_channel1_2.ReadMem(MB_MDE_EK, ulSize, &m_pucExtendedMem[extBegin.pulEKArray], piError)) {

			return FALSE;
		}


		if (GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError) != 0x0400) {

			// Auf naechsten freien Speicher
			extBegin.pulFreeScope += GetMDEEKFlashSize();

			// Dieses sollte nicht sein!, fehlende MDE - Groesse.
			if (ulSize == (EXTENDED_SIZE_SU - extBegin.pulEKArray)) {

				*piError = IV_ERROR_FILE_FORMAT;
				return FALSE;
			}

			// Da die Tabellendefinitionen fuer die Zutrittskontrolle eine
			// feste Struktur haben, wird aufgrund des 0 - Zeigers im Strukt
			// und des Flags fuer Auslesen, im Setup das Einfuegen vorgenommen.
			extBegin.pulEKArrayEntrance = extBegin.pulFreeScope;

			// Groesse der Zutrittskontrolle - Eingabeketten bereits im Eep?
			ulTmp = *(ULONG*)&m_pucEepMem[P_ENTRANCE_EK_SIZE];
			ChangeOrderLong((unsigned char*)&ulTmp);
			// Liegt im moeglichen Bereich?
			if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0)) {

				ulSize = ulTmp;
			} else {
			
				*piError = IV_ERROR_FILE_FORMAT;
				return FALSE;
			}

			// Zutritts - Eingabeketten - lesen
			if (!m_channel1_2.ReadMem(MB_ENTRANCE_EK, ulSize, &m_pucExtendedMem[extBegin.pulEKArrayEntrance], piError)) {

				return FALSE;
			}

			// Auf naechsten freien Speicher
			extBegin.pulFreeScope += GetEntranceEKFlashSize();

			// Der Zeiger auf den naechsten freien Speicherplatz in der
			// erweiterten Struktur muss aufgrund der Abwaertskompatibilitaet
			// auf 0 gesetzt bleiben.
			*(TExtendedFlash*)m_pucExtendedMem = extBegin;
		} else {

			// Auf naechsten freien Speicher
			extBegin.pulFreeScope += GetMDEEKFlashSize_0400();
		}

		temp = 0; // Bedienstopp ruecksetzen
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}
		res = TRUE;
		break;}

	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		res = FALSE;
		break;

	case 4: // PZEMasterIV mit Protokollversion 4

		res = DownLoadSetup4(GetDeviceIDRemote());
		if (res < 0) {

			// Kommunikationsfehler
			res = -res;
			// Fehlercode umsetzen
			*piError = res;
			res = 0;
		} else if (res > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((res >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}

			// Fehler
			res = 0;
		} else {

			// OK
			res = 1;
		}
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
		res = 0;
	}

	if (res) {

		// EEPROM - schreiben
		fwrite(m_pucEepMem, sizeof(unsigned char), EEP_SIZE, pFile);

		// Flashbereich - schreiben
		fwrite(m_pucFlashMem, sizeof(unsigned char), FLASH_SIZE_SU, pFile);
		
		// Leere Filelist schreiben
		unsigned char buf[MAX_DATENSATZ_BESCHREIBUNG];
		memset(buf, 0, MAX_DATENSATZ_BESCHREIBUNG);

		fwrite(buf, sizeof(unsigned char), MAX_DATENSATZ_BESCHREIBUNG, pFile);

		// Konfiguration - schreiben
		fwrite(m_pucConfigMem, sizeof(unsigned char), CONFIG_SIZE_SU, pFile);

		// Erweiterter Flashbereich - schreiben
		fwrite(m_pucExtendedMem, sizeof(unsigned char), EXTENDED_SIZE_SU, pFile);

		// Leere Filelist fuer Zutrittskontrolle (Version 1)
		fwrite(buf, sizeof(unsigned char), 4, pFile);

		// Leere Filelist fuer Zutrittskontrolle (Version 2)
		fwrite(buf, sizeof(unsigned char), MAX_DATENSATZ_BESCHREIBUNG, pFile);
	} 
		
	if (pFile) {
	
		// Datei schlieszen
		fclose(pFile);
	}

	return res;
}

int CAEChannel::DownLoadSetup4(short deviceIDRemote)
{
	int i, res = IV_ERROR_SUCCESS;
	unsigned char buf[0x400];
	unsigned long sizeSetup = 0;
	unsigned char countTable, countField;
	unsigned long sizeAccess;


	//
	unsigned long address, size, crc, crcTest = 0;
	address = size = crc = 0;
	unsigned char *data = 0;

	// Segmentbegin: Bedienung stoppen
	if ((res = m_channel4.BeginGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc)) != IV_ERROR_SUCCESS) {

		if ((res & 0xFF) == 1) {
			// Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
			// Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
			res = -IV_ERROR_FUNCTION_UNAVAILABLE;
		} else {

			return res;
		}
	}

	
	//////////////////////////////
	// Begin

	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Vorhandene zur moeglichen Groesse pruefen
	if (size < (EEP_SIZE_SU + 0x80)) {

		return -IV_ERROR_PROCESSING;
	}

	//////////////////////////////
	// Grunddaten
	
	// EEPROM uebertragen
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_pucEepMem, EEP_SIZE_SU)) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += EEP_SIZE_SU;
	// CONFIG uebertragen (128 Byte)
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_pucConfigMem, 0x80)) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += 0x80;

	// Groesse der Flashdaten fuer Setup aus EepMem entnehmen
	ChangeOrderLong(&m_pucEepMem[P_FLASH_SIZE_SU]);
	sizeSetup = *(unsigned long*)&m_pucEepMem[P_FLASH_SIZE_SU];
	ChangeOrderLong(&m_pucEepMem[P_FLASH_SIZE_SU]);

	// FLASH uebertragen
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_pucFlashMem, sizeSetup)) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += sizeSetup;


	unsigned long sizeEK;

	// Groesse der Zutrittskontroll - Eingabeketten aus EepMem entnehmen.
	ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);
	sizeEK = *(unsigned long*)&m_pucEepMem[P_ENTRANCE_EK_SIZE];
	ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);

	// Zeigeroffsets fuer Firmware anpassen
	if (sizeEK > sizeof(buf)) {
		// Puffer zu klein, also kein Zutritt
		return -IV_ERROR_FILE_FORMAT;
	}
	memset(buf, 0, sizeof(buf));

	//////////////////////////////
	// Pointerarray + Strukturen der Zutrittslisten uebertragen
	
	// Eingabekette der Zutrittskontrolle
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, buf, sizeEK)) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += sizeEK;

	// Pointerarray lesen
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer))) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += sizeof(m_entrance2Pointer);

	///////////////
	// ZK II

	// Anzahl der Zutrittslisten ermitteln
	for (i=0, countTable=0; countTable < (sizeof(m_entrance2Pointer) / sizeof(unsigned long)); countTable++) {

		if (m_entrance2Pointer[countTable] == 0) {

			break;
		}

		// Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
		i += m_entrance2Table[countTable].count;
	}
	sizeAccess = sizeof(m_entrance2Pointer);
	sizeAccess += countTable * sizeof(STable2);
	sizeAccess += i * sizeof(STable2Field);

	// Strukturen lesen
	for (i=0; i < countTable; i++) {

		// Tabellenstruktur uebertragen
		if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)&m_entrance2Table[i], sizeof(STable2))) != IV_ERROR_SUCCESS) {

			return res;
		}
		address += sizeof(STable2);

		// Feldstrukturen lesen
		if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2TableField[i], m_entrance2Table[i].count * sizeof(STable2Field))) != IV_ERROR_SUCCESS) {

			return res;
		}
		address += m_entrance2Table[i].count * sizeof(STable2Field);
	}


	///////////////
	// Grunddaten

	// CRC berechnen
	crcTest = 0xFFFFFFFF;
	crcTest = m_channel4.GetCRC32(m_pucEepMem, EEP_SIZE_SU, crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(m_pucConfigMem, 0x80, crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(m_pucFlashMem, sizeSetup, crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(buf, sizeEK, crcTest, FALSE);


	// CRC der Zutrittslisten ermitteln
	crcTest = m_channel4.GetCRC32((unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer), crcTest, FALSE);

	for (i=0; i < countTable; i++) {

		crcTest = m_channel4.GetCRC32((unsigned char*)&m_entrance2Table[i], sizeof(STable2), crcTest, FALSE);

		for (countField=0; countField < m_entrance2Table[i].count; countField++) {

			// Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
			crcTest = m_channel4.GetCRC32((unsigned char*)&m_entrance2TableField[i][countField], sizeof(STable2Field), crcTest, ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))));
		}
	}
	
	//////////////////////////////
	// Ende

	if (crc != crcTest) {

		m_channel4.Trace("ERROR(0x%08X, 0x%08X)\n", crc, crcTest);
	}

	// Segmentend: Setup uebertragen
	if ((res = m_channel4.EndGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP)) != IV_ERROR_SUCCESS) {

		return res;
	}

	//////////////////////////////
	// Begin
	address = size = crc = 0;
	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	//////////////////////////////
	// Extended - Speicher Lesen

	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_pucExtendedMem, size)) != IV_ERROR_SUCCESS) {

		return res;
	}
	address += size;

	///////////////
	// Extended - Speicher pruefen

	// Extended - Speicher
	crcTest = 0xFFFFFFFF;
	crcTest = m_channel4.GetCRC32(m_pucExtendedMem, size, crcTest, TRUE);

	if (crc != crcTest) {

		m_channel4.Trace("ERROR(0x%08X, 0x%08X)\n", crc, crcTest);
	}

	//////////////////////////////
	// Ende

	// Segmentend: Setup uebertragen
	if ((res = m_channel4.EndGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Die Eingabekettenzeiger anpassen, da diese fuer die FW vor dem Einspielen angepasst wurden.
	// Zeigeroffsets anpassen (addieren von Offset)
	AdjustExtendedMemPointer();

	crc = 0;
	// Segmentend: Bedienung stoppen
	if ((res = m_channel4.EndGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_UPDATE_BG)) != IV_ERROR_SUCCESS) {

		return res;
	}

/*
	// GV setzen?
	if (extBegin.pulValueGVx) {
	
		SGlobVar *globVar = (SGlobVar*)m_pucFlashMem;
		char name[64];
		memset(name, 0, sizeof(name));

		// Struktur lesen
		TValueGVx ValueGVx = *(TValueGVx*)&m_pucExtendedMem[extBegin.pulValueGVx];

		for (i=0; i < 8; i++) {

			if (ValueGVx.ucFlags & (1 << i)) {
			
				sprintf(name, "SETUP.%s", &globVar->SFeld[i]);
				if ((res = m_channel4.SetVariable(deviceIDRemote, name, (char*)&ValueGVx.szValue[i])) != IV_ERROR_SUCCESS) {
				
					// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
					// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
					if (res == 3) {
						// Fehler liegt bei Wert
						err = IV_ERROR_VARIABLE_VALUE;
					} else if (res == 2) {
						// Variable nicht bekannt
						err = IV_ERROR_VARIABLE_UNDEFINED;
					}

					return res;
				}
			}
		}
	}	

	// Standardwerte fuer Systemvariable
	//############################
	// 2005-09-19: Erweiterte Struktur fuer Parameter
	if (((m_pucEepMem[P_SETUP_VERSION    ] == 3) &&
		 (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
		 ((m_pucEepMem[P_SETUP_VERSION + 2] > 34) ||
		  (m_pucEepMem[P_SETUP_VERSION + 2] == 34) &&
		  (m_pucEepMem[P_SETUP_VERSION + 3] >  5)) ) ||

		 ((m_pucEepMem[P_SETUP_VERSION   ] == 4) &&
		  (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
		  ((m_pucEepMem[P_SETUP_VERSION + 2] > 0) || 
		   (m_pucEepMem[P_SETUP_VERSION + 2] == 0) &&
		   (m_pucEepMem[P_SETUP_VERSION + 3] > 2))) ) {

		TExtendedFlash extBegin = *(TExtendedFlash*)m_pucExtendedMem;

		// Liegt die Struktur vor?
		if (extBegin.pulExtended) {

			TExtended extParam = *(TExtended*)&m_pucExtendedMem[extBegin.pulExtended];

			char value[32], name[64];
			// Buchungswiederholsperrdauer
			sprintf(name, "_SERVICE.DBLINPUT");
			sprintf(value, "%i", extParam.doubleInputTime);
			// 
			if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, value)) != IV_ERROR_SUCCESS) {
			
				// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
				// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
				if (res == 3) {
					// Fehler liegt bei Wert
					err = IV_ERROR_VARIABLE_VALUE;
					res = IV_ERROR_SUCCESS;
				} else if (res == 2) {
					// Variable nicht bekannt
					err = IV_ERROR_VARIABLE_UNDEFINED;
					res = IV_ERROR_SUCCESS;
				}

				return res;
			}
			// Abblendedauer fuer Bestaetigungsfeld
			sprintf(name, "_SERVICE.DEMANDTIMER");
			sprintf(value, "%i", extParam.confirmationTime);
			// 
			if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, value)) != IV_ERROR_SUCCESS) {
			
				// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
				// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
				if (res == 3) {
					// Fehler liegt bei Wert
					err = IV_ERROR_VARIABLE_VALUE;
					res = IV_ERROR_SUCCESS;
				} else if (res == 2) {
					// Variable nicht bekannt
					err = IV_ERROR_VARIABLE_UNDEFINED;
					res = IV_ERROR_SUCCESS;
				}

				return res;
			}
		}
	}
*/
	return res;
}

int CAEChannel::CheckDevice(short deviceIDRemote, int *piError, int iPollRetry)
{
	int iRes = IV_ERROR_SUCCESS;

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:

		for (m_iDevicePollRetry=1; m_iDevicePollRetry <= iPollRetry; m_iDevicePollRetry++) {

			int tmp;
			if (m_channel1_2.ReadMem(0ul, 2, (unsigned char*)&tmp, piError)) {

				return TRUE;
			}

			Sleep(20);
		}
		break;

	case 3: { // PZEMasterIV mit Protokollversion 3
		char szVersion[32];
		iRes = m_channel3.GetVersion(GetDeviceIDRemote(), szVersion);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	case 4: { // PZEMasterIV mit Protokollversion 4
		char szVersion[32];
		iRes = m_channel4.GetVersion(GetDeviceIDRemote(), szVersion);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	// Letzter Fehler ist aktueller
	return FALSE;
}

int CAEChannel::Restart(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int tmp = 0;

	int iRes = IV_ERROR_SUCCESS;
	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 2:

		if (m_channel1_2.WriteMem(MB_RESTART, 1, (unsigned char*)&tmp, piError)) {

			return TRUE;
		}
		return FALSE;

	case 1:
	case 3:
	case 4:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	// Letzter Fehler ist aktueller
	return FALSE;
}

int CAEChannel::GetProtocolVersion2(short deviceIDRemote, unsigned char detailed, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	unsigned char Buf[4];
	
	// Protokollverwendung pruefen
	unsigned char pv = GetProtocolVersion(GetDeviceIDRemote());
	switch(pv) {
	case 1:
	case 2:

		// Detailierte Version auslesen.
		if (detailed) {

			// Version fuer Setup lesen
			if (!m_channel1_2.ReadMem(MB_GET_VERSION, 2, Buf, piError)) {

				return FALSE;
			}
		} else {

			if (pv == 1) {

				// Es wird angenommen ein AEIII+ mit neuester Com-Version
				Buf[0] = 0x04;
				Buf[1] = 0x02;
			} else {

				// Es wird angenommen ein PZE Master IV mit neuester Com-Version
				Buf[0] = 0x05;
				Buf[1] = 0x01;
			}
		}
		break;

	case 3: 
		// Ab dem Protokolltyp 3 wird nur noch ueber die Firmwareversion entschieden.
		Buf[0] = 0x06;
		Buf[1] = 0x00;
		break;

	case 4:
		// Ab dem Protokolltyp 4 wird nur noch ueber die Firmwareversion entschieden.
		Buf[0] = 0x07;
		Buf[1] = 0x00;
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return FALSE;
	}
	

	// Version fuer Kommunikation
	
	// Ab Firmware 04.01.00 Protokollanpassung (Version 4) und Listen in Tabellenart abhandeln.
	if ((Buf[0] == 7 && Buf[1] == 0) ||
		// Ab Firmware 04.01.00 Protokollanpassung (Version 3)
		(Buf[0] == 6 && Buf[1] == 0) ||
		// Ab Firmware 04.00.23 Listengroesse auf 256kByte erhoeht.
		(Buf[0] == 5 && Buf[1] == 2) ||
		// Ab Firmware 04.00.21 Listen im Hintergrund laden
		(Buf[0] == 5 && Buf[1] == 1) ||
		// Ab Firmware 04.00.19 des PZEMasterIV (Version 2)
		(Buf[0] == 5 && Buf[1] == 0) ||
		// Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
		(Buf[0] == 4 && Buf[1] == 2) ||
		// Ab Firmware 03.00.12 mit Zutrittskontrolle
		(Buf[0] == 4 && Buf[1] == 1) ||
	    // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
		(Buf[0] == 4 && Buf[1] == 0) ) { // (Version 1)

		// Version als Rueckgabe
		return (Buf[0] << 8) | Buf[1];
	}

	*piError = IV_ERROR_VERSION_FIRMWARE;
	return FALSE;
}

int CAEChannel::GetSerialNum(short deviceIDRemote, int *piSerialNum, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: 
	case 2: {
		unsigned char buf[10];

		// Seriennummer und Optionsbyte lesen
		if (!m_channel1_2.ReadMem(MB_SERIEN_MEMORY, 4, buf, piError)) {

			return FALSE;
		}

		*piSerialNum = (int)(buf[2] << 8);
		*piSerialNum |= (int)(buf[3]);
		return TRUE;}

	case 3: {
		int iRes;
		unsigned long serial;
		iRes = m_channel3.GetSerial(GetDeviceIDRemote(), serial);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			*piSerialNum = serial;
			iRes = 1;
		}

		return iRes;}

	case 4: {
		int iRes;
		unsigned long serial;
		iRes = m_channel4.GetSerial(GetDeviceIDRemote(), serial);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			*piSerialNum = serial;
			iRes = 1;
		}

		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// ListenBuffer in AE laden
int CAEChannel::LoadListData(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int i;
	// Listengesamtgroesse berechnen und ggf. sortieren.
	unsigned long listSize = MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned long);
	unsigned long listPointer[MAX_DATENSATZ_BESCHREIBUNG];
	for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

		if (m_listSize[i]) {
			
			listPointer[i] = listSize + FLASH_LIST_START;
			ChangeOrderLong((unsigned char*)&listPointer[i]);

			listSize += m_listSize[i];

		} else {

			listPointer[i] = 0;
		}
	}

	if (listSize == (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned long))) {

		*piError = IV_ERROR_LIST_PROCESSING;
		return FALSE; // keine Listen vorhanden
	}


	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: 
	case 2: {

		unsigned long listStart = FLASH_LIST_START;
		unsigned char temp = 1; // Bedienstopp setzen
		// Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
		// Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
		// dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
		switch(GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError)) {
		case 0: // Kommunikationsfehler
			return FALSE;

		case 0x0501:
			// Listengroesse nur maximal 100kByte
			if (listSize >= FLASH_SIZE_LIST) {

				*piError = IV_ERROR_LIST_PROCESSING;
				return FALSE;
			}
			break;

		case 0x0502:
			listStart = FLASH_LIST_START_NEW;

			// Listengroesse nur maximal 244kByte
			if (listSize >= FLASH_SIZE_LIST_NEW) {

				*piError = IV_ERROR_LIST_PROCESSING;
				return FALSE;
			}
			break;
		/*
		case 0x0503:
			// 2004-11-26
			//@@: Listenfunktion fuer Listentest umleiten
			return LoadNewListData(deviceIDRemote, piError);
			break;
		*/

		default:
			// Listengroesse nur maximal 100kByte
			if (listSize >= FLASH_SIZE_LIST) {

				*piError = IV_ERROR_LIST_PROCESSING;
				return FALSE;
			}
			if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

				return FALSE;
			}
		}

		// Flash schreiben 
		if (!m_channel1_2.WriteMem(listStart, (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned long)), (unsigned char*)listPointer, piError)) {

			return FALSE;
		}

		unsigned long tmpSize = 0;
		for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

			if (m_listSize[i]) {

				// Datensatzanzahl drehen
				ChangeOrderLong((unsigned char*)&m_pucListMem[i][0]);

				// Flash schreiben 
				if (!m_channel1_2.WriteMem(listStart + (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned long)) + tmpSize, m_listSize[i], m_pucListMem[i], piError)) {

					// Datensatzanzahl drehen
					ChangeOrderLong((unsigned char*)&m_pucListMem[i][0]);
					return FALSE;
				}

				// Datensatzanzahl drehen
				ChangeOrderLong((unsigned char*)&m_pucListMem[i][0]);
				tmpSize += m_listSize[i];
			}
		}

		temp = 0; // Bedienstopp ruecksetzen
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}	

		return TRUE;}
		
	case 3: // PZEMasterIV mit Protokollversion 3

		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes = LoadListData4(GetDeviceIDRemote());
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// Fuer neue Listentechnik
int CAEChannel::LoadListData4(short deviceIDRemote)
{
	unsigned char tableNum;
	unsigned long res = IV_ERROR_SUCCESS;

	STable2 table;
	char name[32];

	//
	unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
	unsigned long address, size, crc;
	address = size = crc = 0;

	// Systemstopp Listen uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

		// Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
		if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG)) {

			segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
			continue;
		}

		return res;
	}

	// Listendaten Tabellenweise uebertragen
	for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++) {

		// Nicht importierte Listen uebergehen
		if (m_listSize[tableNum] == 0) continue;

		// Maximale Segmentanzahl nicht ueberschreiten
		if (tableNum && !(tableNum % 16)) {

			crc = 0;
			// Systemstopp Listen uebertragen
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

				return res;
			}

			// Systemstopp Listen uebertragen
			if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}

		memset(name, 0, sizeof(name));
		sprintf(name, "LIST.%i", tableNum);
		// Zur Liste Informationen ermitteln
		if ((res = m_channel4.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS) {

			return res;
		}
		strcpy(name, table.name);
		sprintf(table.name, "LIST.%s", name);

		// Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
		if (m_listSortCheck[tableNum] == false) {

			int fieldOffs = 0;

			if (table.flags & TF_SORT && table.sort < table.count) {

				char name[64];
				int fieldNum;
				STable2Field field;
				memset(&field, 0, sizeof(STable2Field));
				m_ucListRecordSize[tableNum] = 0;
				// Alle Felder fuer Offset lesen
				for (fieldNum=0; fieldNum < table.count; fieldNum++) {

					if (fieldNum < (table.sort + 1)) {

						// Datensatzoffset fuer Sortierfeld aufsummieren.
						fieldOffs += field.size;
					}

					sprintf(name, "LIST.%i.%i", tableNum, fieldNum);

					if ((res = m_channel4.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS) {

						if (res < 0) {

							// Kommunikationsfehler
							return res;

						} else {

							// Funktionsfehler
							return res;
						}
					}

					// Datensatzgroesse fuer Swap-Routine berechnen
					m_ucListRecordSize[tableNum] += field.size;
				}
/*
				int x, y;
				for (x=0; x < 10 && x <= *(unsigned long*)&m_pucListMem[tableNum]; x++) {

					TRACE("Unsort(%i.%i): ", tableNum, x+1);
					for (y=0; y < m_ucListRecordSize[tableNum] + sizeof(unsigned long); y++) {

						if (m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned long) + y]) {

							TRACE("%c", m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned long) + y]);
						} else {

							TRACE(" ");
						}
					}
					TRACE("\n");
				}
*/
				// Liste sortieren.
				QuicksortListRecord(tableNum, fieldOffs, 0, *(unsigned long*)m_pucListMem[tableNum] - 1);
/*
				for (x=0; x < 10 && x <= *(unsigned long*)&m_pucListMem[tableNum]; x++) {

					TRACE("Sort(%i.%i): ", tableNum, x+1);
					for (y=0; y < m_ucListRecordSize[tableNum] + sizeof(unsigned long); y++) {

						if (m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned long) + y]) {

							TRACE("%c", m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned long) + y]);
						} else {

							TRACE(" ");
						}
					}
					TRACE("\n");
				}
*/
				// Bei allen Listen ist 0 die sortierte Spalte
				// Flag dafuer dass die Liste sortiert wurde setzen.
				table.flags |= TF_SORTED;
			}

			// Sortiert oder zumindes geprueft.
			m_listSortCheck[tableNum] = true;
		}

		// CRC berechnen
		crc = 0xFFFFFFFF;
		crc = m_channel4.GetCRC32((unsigned char*)&table, sizeof(STable2), crc, FALSE);
		crc = m_channel4.GetCRC32(m_pucListMem[tableNum], m_listSize[tableNum], crc);

		// Groesze berechnen
		size = sizeof(STable2) + m_listSize[tableNum];

		// Systemstopp Liste uebertragen
		if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_LIST, address, size, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Daten uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&table, sizeof(STable2))) != IV_ERROR_SUCCESS) {

			return res;
		}
		// Daten uebertragen
		address += sizeof(STable2);
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucListMem[tableNum], m_listSize[tableNum])) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Systemstopp Liste uebertragen
		if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_LIST, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}
	}

	crc = 0;
	// Systemstopp Liste uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntranceListData(short deviceIDRemote, int iListNum, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	// Alle Listen laden?
	if (iListNum == -1) {

		// Mindestens eine Liste geladen?
		int i;
		for (i=0; i < 4; i++) {
			if (m_ulEntranceListSize[i] != 0) {

				break;
			}
		}
		if (i == 4) {

			*piError = IV_ERROR_LIST_PROCESSING;
			return FALSE;
		}
	} else {

		if (m_ulEntranceListSize[iListNum] == 0) {

			*piError = IV_ERROR_LIST_PROCESSING ;
			return FALSE; // keine Listen vorhanden
		}
	}

	// Protokollklasse
	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 3:
	case 4:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;
	case 1:
	case 2:
		break;
	default:
		*piError = IV_ERROR_UNREACHABLE;
		return FALSE;
	}

	unsigned char temp = 1; // Bedienstopp setzen
	// Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
	// Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
	// dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
	switch(GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError)) {
	case 0x0501:
	case 0x0502:
		break;

	default:
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}
	}

	switch(iListNum) {
	case -1:
	case 0:
		if (m_ulEntranceListSize[0] != 0) {
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_0, m_ulCompressedListSize[0], m_pucCompressedListMem[0], piError, 32)) {

				return FALSE;
			}
		}
		if (iListNum != -1) break;
	case 1:
		if (m_ulEntranceListSize[1] != 0) {
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_1, m_ulCompressedListSize[1], m_pucCompressedListMem[1], piError, 32)) {

				return FALSE;
			}
		}
		if (iListNum != -1) break;
	case 2:
		if (m_ulEntranceListSize[2] != 0) {
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_2, m_ulCompressedListSize[2], m_pucCompressedListMem[2], piError, 32)) {

				return FALSE;
			}
		}
		if (iListNum != -1) break;
	case 3:
		if (m_ulEntranceListSize[3] != 0) {
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_3, m_ulCompressedListSize[3], m_pucCompressedListMem[3], piError, 32)) {

				return FALSE;
			}
		}
	}

	temp = 0; // Bedienstopp ruecksetzen
	if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

		return FALSE;
	}

	return TRUE;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntrance2ListData(short deviceIDRemote, int iListNum, int *piError)
{
	int iError = IV_ERROR_SUCCESS;

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return FALSE;
	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;

	case 4: {
		// Listendaten zum AEIV uebertragen
		if ((iError = LoadEntrance2ListData4(deviceIDRemote, iListNum)) != IV_ERROR_SUCCESS) {
			
			if (iError < 0) {

				// Kommunikationsfehler
				iError = -iError;
				// Fehlercode umsetzen
				*piError = iError;
			} else if (iError > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iError >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
			}

			return FALSE;
		}
		return TRUE;}
	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntrance2ListData4(short deviceIDRemote, unsigned char listNum)
{
	// Alle Listen laden?
	if (listNum == 0xFF) {

		// Mindestens eine Liste geladen?
		int i;
		for (i=0; i < 7; i++) {
			if (m_ulEntrance2ListSize[i] != 0) {

				break;
			}
		}
		if (i == 7) {

			return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
		}
	} else {

		if (m_ulEntrance2ListSize[listNum] == 0) {

			return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
		}
	}

	unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
	int res = IV_ERROR_SUCCESS;
	unsigned long address, size, crc;
	address = size = crc = 0;

	// Systemstopp Setup uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

		// Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
		if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG)) {

			segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
			continue;
		}

		return res;
	}

	switch(listNum) {
	case 0xFF:
	case 0x00:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[0] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 0)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x01:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[1] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 1)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x02:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[2] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 2)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x03:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[3] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 3)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x04:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[4] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 4)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x05:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[5] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 5)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x06:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[6] != 0) {

			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 6)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
	}

	crc = 0;
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntrance2List4(short deviceIDRemote, unsigned char listNum)
{
	int res = IV_ERROR_SUCCESS;
	STable2 table;
	char name[32];
	memset(name, 0, sizeof(name));
	sprintf(name, "ACCESS.%i", listNum);
	// Zur Liste Informationen ermitteln
	if ((res = m_channel4.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS) {

		return res;
	}
	strcpy(name, table.name);
	sprintf(table.name, "ACCESS.%s", name);

	// Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
	if (m_entrance2SortCheck[listNum] == false) {
		
		int fieldOffs = 0;
		
		if (table.flags & TF_SORT && table.sort < table.count) {
			
			//char name[64];
			int fieldNum;
			STable2Field field;
			memset(&field, 0, sizeof(STable2Field));
			// Alle Felder fuer Offset lesen
			for (fieldNum=0; fieldNum < table.count; fieldNum++) {
				
				if (fieldNum < (table.sort + 1)) {
					
					// Datensatzoffset fuer Sortierfeld aufsummieren.
					fieldOffs += field.size;
				}
				// Da die Datensatzgroesse bekannt ist, brauchen die restlichen Felder nicht ermittelt zu werden. 
				else { break; }
				
				/*
				sprintf(name, "ACCESS.%i.%i", listNum, fieldNum);
				
				if ((res = m_channel4.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS) {
					
					if (res < 0) {
						
						// Kommunikationsfehler
						return res;
						
					} else {
						
						// Funktionsfehler
						return res;
					}
				}
				*/
				field = m_entrance2TableField[listNum][fieldNum];
				
				// Datensatzgroesse wird fuer Swap-Routine benoetigt!
			}
			
			// Liste sortieren.
			QuicksortEntrance2Record(listNum, fieldOffs, 0, *(unsigned long*)m_pucEntrance2ListMem[listNum] - 1);
			
			// Bei allen Listen ist 0 die sortierte Spalte
			// Flag dafuer dass die Liste sortiert wurde setzen.
			table.flags |= TF_SORTED;
		}
		
		// Sortiert oder zumindes geprueft.
		m_entrance2SortCheck[listNum] = true;
	}
	

	//
	unsigned short art = IV_VALUE_SEGMENT_LIST_ACCESS;
	unsigned long address, size, crc;
	address = size = crc = 0;
/*
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentBegin(deviceIDRemote, SS_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
*/
	// CRC berechnen
	crc = 0xFFFFFFFF;
	crc = m_channel4.GetCRC32((unsigned char*)&table, sizeof(STable2), crc, FALSE);
	crc = m_channel4.GetCRC32(m_pucEntrance2ListMem[listNum], m_ulEntrance2ListSize[listNum], crc);

	// Groesze berechnen
	size = sizeof(STable2) + m_ulEntrance2ListSize[listNum];

	// Systemstopp List uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, art, address, size, crc)) != IV_ERROR_SUCCESS) {

		// Typ unbekannt?, dann als "normale" Liste versuchen..
		if (((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS)) {

			art = IV_VALUE_SEGMENT_LIST;
			continue;
		}

		return res;
	}

	// Daten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&table, sizeof(STable2))) != IV_ERROR_SUCCESS) {

		return res;
	}
	// Daten uebertragen
	address += sizeof(STable2);
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucEntrance2ListMem[listNum], m_ulEntrance2ListSize[listNum])) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Systemstopp Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, art, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
/*
	crc = 0;
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentEnd(deviceIDRemote, SS_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
*/
	return res;
}
/*
// ListenBuffer in AE laden
int CAEChannel::DownloadLists(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	char fileName[MAX_PATH];
	UCHAR sizes[MAX_DATENSATZ_BESCHREIBUNG];
	memset(sizes, 0, sizeof(sizes));

	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 3:
	case 4:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return FALSE;

	case 1: 
	case 2:
		break;

	default:
		*piError = IV_ERROR_RECEIVE;
		return FALSE;
	}

	if (!m_channel1_2.SetDeviceIDRemote((unsigned char)deviceIDRemote)) {

		*piError = IV_ERROR_DEVICE_ID_REMOTE;
		return FALSE;
	}

	// Listenpointer lesen
	if (!m_channel1_2.ReadMem(FLASH_LIST_START, (MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG)), m_pucListMem, piError)) {

		return FALSE;
	}

	int i;
	for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

		if (!*(ULONG*)&m_pucListMem[i * sizeof(ULONG)]) {

			break;
		} else {

			ChangeOrderLong(&m_pucListMem[i * sizeof(ULONG)]);
		}
	}
	// Wenn mindestens eine Liste da ist
	if (i) {

		i--;
		// Anzahl Datensaetze lesen
		if (!m_channel1_2.ReadMem(*(ULONG*)&m_pucListMem[i * sizeof(ULONG)], sizeof(ULONG), 
			&m_pucListMem[*(ULONG*)&m_pucListMem[i * sizeof(ULONG)] - FLASH_LIST_START], piError)) {

			return FALSE;
		} else {

			ChangeOrderLong(&m_pucListMem[*(ULONG*)&m_pucListMem[i * sizeof(ULONG)] - FLASH_LIST_START]);
		}

		// Groesze der Listen berechnen
		int numList = 0, numFeld;
		for (; numList <= i; numList++) {

			for (numFeld=0; numFeld < ListenBeschreibung[numList].Felder; numFeld++) {

				sizes[numList] += ListenFeld[numList][numFeld].Cnt;
			}
		}
		// Listenblockgroesze bestimmen
		m_ulListSize = (*(ULONG*)&m_pucListMem[i * sizeof(ULONG)] - FLASH_LIST_START) + sizeof(ULONG) + *(ULONG*)&m_pucListMem[*(ULONG*)&m_pucListMem[i * sizeof(ULONG)] - FLASH_LIST_START] * sizes[i];

		// Listendaten lesen
		if (!m_channel1_2.ReadMem(FLASH_LIST_START + (MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG)),m_ulListSize , &m_pucListMem[(MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG))], piError)) {

			return FALSE;
		}
		// Anzahlangabe drehen
		for (numList=0; numList <= i; numList++) {

			ChangeOrderLong(&m_pucListMem[*(ULONG*)&m_pucListMem[numList * sizeof(ULONG)] - FLASH_LIST_START]);
		}

		// Listendaten ausgeben
		unsigned long numRecord;
		unsigned long offs;
		FILE *fp;
		for (numList=0; numList <= i; numList++) {

			sprintf(fileName, "%s.txt", ListenBeschreibung[numList].Name);
			if ((fp = fopen(fileName, "w")) != NULL) {

				for (numRecord=0; numRecord < *(ULONG*)&m_pucListMem[*(ULONG*)&m_pucListMem[numList * sizeof(ULONG)] - FLASH_LIST_START]; numRecord++) {

					offs = 0;
					for (numFeld=0; numFeld < ListenBeschreibung[numList].Felder; numFeld++) {

						fprintf(fp, "%s%s", &m_pucListMem[(*(ULONG*)&m_pucListMem[numList * sizeof(ULONG)] - FLASH_LIST_START) + sizeof(ULONG) + numRecord * sizes[numList] + offs], (numFeld < ListenBeschreibung[numList].Felder - 1) ? "\t" : "");
						offs += ListenFeld[numList][numFeld].Cnt;
					}
					fprintf(fp, "\n");
				}
				fprintf(fp, "\n");
			}

			fclose(fp);
		}
	}

	return TRUE;
}
*/

// ListenBuffer in AE laden
int CAEChannel::LoadTimeboyListData(short deviceIDRemote, unsigned char id, unsigned char num, int reserve, int *piError)
{
	int iError = IV_ERROR_SUCCESS;

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return FALSE;
	case 3:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return FALSE;

	case 4: {
		// Listendaten zum AEIV uebertragen
		if ((iError = LoadTimeboyListData4(deviceIDRemote, id, num, reserve)) != IV_ERROR_SUCCESS) {
			
			if (iError < 0) {

				// Kommunikationsfehler
				iError = -iError;
				// Fehlercode umsetzen
				*piError = iError;
			} else if (iError > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iError >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
			}

			return FALSE;
		}
		return TRUE;}
	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// ListenBuffer in AE laden
int CAEChannel::LoadTimeboyListData4(short deviceIDRemote, unsigned char id, unsigned char num, int reserve)
{
	// Alle Listen laden?
	if (id == 0xFF) {

		int groupID, listNum, found = 0;
		for (groupID=0; groupID < MAX_DATENSATZ_BESCHREIBUNG; groupID++) {
			
			if (num == 0xFF) {

				// Mindestens eine Liste geladen?
				for (listNum=0; listNum < MAX_DATENSATZ_BESCHREIBUNG; listNum++) {
					if (m_ulTimeboyListSize[groupID][listNum] != 0) {

						found = 1;
						break;
					}
				}
				if (found) {

					break;
				}
			} else {

				// Mindestens eine Liste geladen?
				if (m_ulTimeboyListSize[groupID][num] == 0) {

					found = 1;
					break;
				}
			}
		}
		if (found == 0) {

			return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
		}
	} else {

		if (num == 0xFF) {

			int listNum, found = 0;
			// Mindestens eine Liste geladen?
			for (listNum=0; listNum < MAX_DATENSATZ_BESCHREIBUNG; listNum++) {

				if (m_ulTimeboyListSize[id][listNum] != 0) {

					found = 1;
					break;
				}
			}
			if (found == 0) {

				return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
			}
		} else {

			if (m_ulTimeboyListSize[id][num] == 0) {

				return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
			}
		}
	}

	unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
	int res = IV_ERROR_SUCCESS;
	unsigned long address, size, crc;
	address = size = crc = 0;

	// Systemstopp Listen uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

		// Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
		if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG)) {

			segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
			continue;
		}

		return res;
	}

	// Alle ids abarbeiten?
	if (id == 0xFF) {

		int tableNum, tableCount = 0;
		int groupID;
		for (groupID=0; groupID < MAX_DATENSATZ_BESCHREIBUNG; groupID++) {

			if (num == 0xFF) {

				for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++) {

					// nur geladene Liste bei 0xFF uebertragen
					if (m_ulTimeboyListSize[groupID][tableNum] != 0) {

						// Maximale Segmentanzahl nicht ueberschreiten
						if (tableCount && !(tableCount % 16)) {

							crc = 0;
							// Systemstopp Listen uebertragen
							if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

								return res;
							}

							// Systemstopp Listen uebertragen
							if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

								return res;
							}
						}

						// Liste uebertragen
						if ((res = LoadTimeboyList4(deviceIDRemote, groupID, tableNum, reserve)) != IV_ERROR_SUCCESS) {

							return res;
						}

						tableCount++;
					}
				}
			} else {

				// nur geladene Liste bei 0xFF uebertragen
				if (m_ulTimeboyListSize[groupID][num] != 0) {

					// Maximale Segmentanzahl nicht ueberschreiten
					if (tableCount && !(tableCount % 16)) {

						crc = 0;
						// Systemstopp Listen uebertragen
						if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

							return res;
						}

						// Systemstopp Listen uebertragen
						if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

							return res;
						}
					}

					// Liste uebertragen
					if ((res = LoadTimeboyList4(deviceIDRemote, groupID, num, reserve)) != IV_ERROR_SUCCESS) {

						return res;
					}

					tableCount++;
				}
			}
		}
	} else {
		if (num == 0xFF) {

			int tableNum, tableCount = 0;
			for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++) {

				// nur geladene Liste bei 0xFF uebertragen
				if (m_ulTimeboyListSize[id][tableNum] != 0) {

					// Maximale Segmentanzahl nicht ueberschreiten
					if (tableCount && !(tableCount % 16)) {

						crc = 0;
						// Systemstopp Listen uebertragen
						if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

							return res;
						}

						// Systemstopp Listen uebertragen
						if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS) {

							return res;
						}
					}

					// Liste uebertragen
					if ((res = LoadTimeboyList4(deviceIDRemote, id, tableNum, reserve)) != IV_ERROR_SUCCESS) {

						return res;
					}

					tableCount++;
				}
			}
		} else {

			// Liste uebertragen
			if ((res = LoadTimeboyList4(deviceIDRemote, id, num, reserve)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
	}

	crc = 0;
	// Systemstopp Listen uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadTimeboyList4(short deviceIDRemote, unsigned char id, unsigned char num, int reserve)
{
	int res = IV_ERROR_SUCCESS;
	STable2 table;
	memset(&table, 0, sizeof(STable2));
	table.num = num;
	sprintf(table.name, "TBLIST.%i.%i", id, num);
	table.flags = m_timeboyFlags[id][num];

	//
	unsigned short art = IV_VALUE_SEGMENT_LIST_TIMEBOY;
	unsigned long address, size, crc;
	address = size = crc = 0;
/*
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentBegin(deviceIDRemote, SS_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
*/
	// CRC berechnen
	crc = 0xFFFFFFFF;
	crc = m_channel4.GetCRC32((unsigned char*)&table, sizeof(STable2), crc, FALSE);
	crc = m_channel4.GetCRC32(m_pucTimeboyListMem[id][num], m_ulTimeboyListSize[id][num], crc);

	// Groesze berechnen
	size = sizeof(STable2) + m_ulTimeboyListSize[id][num];

	// Systemstopp List uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, art, address, size, crc)) != IV_ERROR_SUCCESS) {

		// Typ unbekannt?, dann als "normale" Liste versuchen..
		if (((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS)) {

			art = IV_VALUE_SEGMENT_LIST;
			continue;
		}

		return res;
	}

	// Daten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&table, sizeof(STable2))) != IV_ERROR_SUCCESS) {

		return res;
	}
	// Daten uebertragen
	address += sizeof(STable2);
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucTimeboyListMem[id][num], m_ulTimeboyListSize[id][num])) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Systemstopp Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, art, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
/*
	crc = 0;
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentEnd(deviceIDRemote, SS_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
*/
	return res;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int CAEChannel::CheckData(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:  // AEIII+
	case 2: { // PZEMasterIV
		unsigned char ucIsData = FALSE;

		// Abfrage ob Daten vorhanden
		if (!m_channel1_2.ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {
			
			return -1;
		}

		return ucIsData;}

	case 3: {// PZEMasterIV mit Protokollversion 3
		int iRes = IV_ERROR_SUCCESS;
		unsigned char buf[0x255];
		unsigned short count = 0;
		unsigned char status = 0;

		iRes = m_channel3.GetRecord(GetDeviceIDRemote(), buf, count, status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = -1;
		} else {

			// OK
			iRes = status;
		}
		return iRes;}

	case 4: {// PZEMasterIV mit Protokollversion 4
		int iRes = IV_ERROR_SUCCESS;
		unsigned char buf[0x255];
		int count = 0;
		unsigned char status = 0;

		iRes = m_channel4.GetRecord(GetDeviceIDRemote(), buf, count, status);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = -1;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = -1;
		} else {

			// OK
			iRes = status;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return -1;
}

// Schliesst Relais mit Nummer fuer Zeit.
int CAEChannel::CloseRelay(short deviceIDRemote, int iNum, int iTime, int *piError)
{
	int iRes = IV_ERROR_SUCCESS;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		return m_channel1_2.CloseRelay((unsigned char)GetDeviceIDRemote(), iNum, iTime, piError);

	case 3: // PZEMasterIV mit Protokollversion 3
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4
		if (iNum > 4) {

			*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			iRes = 0;
		} else {
		
			iRes = m_channel4.SetRelay(GetDeviceIDRemote(), "INTERN", 0, iNum, 1, iTime);
			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				*piError = iRes;
				iRes = 0;
			} else if (iRes > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iRes >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
				// Fehler
				iRes = 0;
			} else {

				// OK
				iRes = 1;
			}
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int CAEChannel::GetRelayState(short deviceIDRemote, int iNum, int *piState, int *piTime, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		return m_channel1_2.GetRelayState((unsigned char)GetDeviceIDRemote(), iNum, piState, piTime, piError);

	case 3: // PZEMasterIV mit Protokollversion 3
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes = IV_ERROR_SUCCESS;
		if (iNum > 4) {

			*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			iRes = 0;
		} else {
		
			unsigned char state;
			unsigned short time;
			iRes = m_channel4.GetRelay(GetDeviceIDRemote(), "INTERN", 0, iNum, state, time);
			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				*piError = iRes;
				iRes = 0;
			} else if (iRes > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iRes >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
				// Fehler
				iRes = 0;
			} else {

				// OK
				*piState = state;
				*piTime = time;
				iRes = 1;
			}
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// Oeffnet Relais mit Nummer.
int CAEChannel::OpenRelay(short deviceIDRemote, int iNum, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		return m_channel1_2.OpenRelay((unsigned char)GetDeviceIDRemote(), iNum, piError);

	case 3: // PZEMasterIV mit Protokollversion 3
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4
		if (iNum > 4) {

			*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			iRes = 0;
		} else {
		
			unsigned short time = 0;
			iRes = m_channel4.SetRelay(GetDeviceIDRemote(), "INTERN", 0, iNum, 0, time);
			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				*piError = iRes;
				iRes = 0;
			} else if (iRes > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iRes >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
				// Fehler
				iRes = 0;
			} else {

				// OK
				iRes = 1;
			}
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// Daten einer Variablen setzen.
int CAEChannel::SetGlobVar(short deviceIDRemote, unsigned char *pucVar, int iType, char *value, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;

	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		return m_channel1_2.SetGlobVar((unsigned char)GetDeviceIDRemote(), pucVar, iType, value, piError);

	case 3: // PZEMasterIV mit Protokollversion 3
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4

		switch(iType) {
		case 0: {
			char name[64];
			memset(name, 0, sizeof(name));
			sprintf(name, "SETUP.%i", *(unsigned char*)pucVar);
			iRes = m_channel4.SetVariable(GetDeviceIDRemote(), name, value);
			break;}
		case 1: {
			char name[64];
			memset(name, 0, sizeof(name));
			strcpy(name, "SETUP.");
			strcat(name, (char*)pucVar);
			iRes = m_channel4.SetVariable(GetDeviceIDRemote(), name, value);
			break;}
		case 2:
			iRes = m_channel4.SetSystemVariable(GetDeviceIDRemote(), (char*)pucVar, value);
			break;
		}
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			switch(iRes) {
			case 1:
			case 2:
				*piError = IV_ERROR_VARIABLE_UNDEFINED;
				break;
			case 3:
				*piError = IV_ERROR_VARIABLE_VALUE;
				break;
			default:
				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iRes >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

// Daten einer Globalen Variablen lesen.
int CAEChannel::GetGlobVar(short deviceIDRemote, unsigned char *pucVar, int iType, char *value, int iValueLength, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	// muss mindestens ein Zeichen anfordern.
	if ((iValueLength - 1) <= 0) {
	
		*piError = IV_ERROR_FUNCTION_PARAMETER;
		return FALSE;
	} 

	int iRes = IV_ERROR_SUCCESS;
	switch(GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2:
		return m_channel1_2.GetGlobVar((unsigned char)GetDeviceIDRemote(), pucVar, iType, value, iValueLength, piError);

	case 3: // PZEMasterIV mit Protokollversion 3
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4
		switch(iType) {
		case 0: {
			char name[64];
			memset(name, 0, sizeof(name));
			sprintf(name, "SETUP.%i", *(unsigned char*)pucVar);
			iRes = m_channel4.GetVariable(deviceIDRemote, name, value);
			break;}
		case 1: {
			char name[64];
			memset(name, 0, sizeof(name));
			strcpy(name, "SETUP.");
			strcat(name, (char*)pucVar);
			iRes = m_channel4.GetVariable(deviceIDRemote, name, value);
			break;}
		case 2:
			iRes = m_channel4.GetSystemVariable(deviceIDRemote, (char*)pucVar, value);
			break;
		}
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			switch(iRes) {
			case 1:
			case 2:
				*piError = IV_ERROR_VARIABLE_UNDEFINED;
				break;
			case 3:
				*piError = IV_ERROR_VARIABLE_VALUE;
				break;
			default:
				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann ist das Geraet zwar aktiv der Befehl
				// wird jedoch nicht unterstuetzt.
				if (((iRes >> 8) & 0xFF) == 2) {

					*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
				} else {

					// Funktionsfehler
					*piError = IV_ERROR_PROCESSING;
				}
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::GetVersionFirmware(short deviceIDRemote, char *buf, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	unsigned char lBuf[32];

	unsigned char pv = GetProtocolVersion(GetDeviceIDRemote());
	switch(pv) {
	case 1: 
	case 2: {

		if (pv == 1) // Nur Protokollversion 1
		switch(GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError)) {
		case 0:
			return FALSE;

		case 0x0400: // AEIII+
		case 0x0401:

			*piError = IV_ERROR_VERSION_FIRMWARE;
			return FALSE;
		}

		// Firmwareversion lesen
		if (!m_channel1_2.ReadMem(MB_VERSION, 32, lBuf, piError)) {

			return FALSE;
		}
		memcpy(buf, lBuf, 32);

		return TRUE;}

	case 3: {
		int iRes = m_channel3.GetVersion(GetDeviceIDRemote(), (char*)lBuf);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
			strcpy(buf, (char*)lBuf);
		}
		return iRes;}

	case 4: {
		int iRes = m_channel4.GetVersion(GetDeviceIDRemote(), (char*)lBuf);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist das Geraet zwar aktiv der Befehl
			// wird jedoch nicht unterstuetzt.
			if (((iRes >> 8) & 0xFF) == 2) {

				*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
			// Fehler
			iRes = 0;
		} else {

			// OK
			iRes = 1;
			strcpy(buf, (char*)lBuf);
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return FALSE;
}

int CAEChannel::GetVersionFirmwareFromFile(char *fileName, char *versionFirmware, int *error)
{
	int res = 0;
	// Lokale Membervariable reseten
	m_fileDevice.Reset();

	if (!(res = m_fileDevice.GetVersionFirmware(fileName, versionFirmware))) {

		switch(m_fileDevice.GetLastErrorState()) {
		case ES_FILE_OPEN:
			*error = IV_ERROR_FILE_OPEN;
			break;

		case ES_HEAD_CRC:
			*error = IV_ERROR_CRC_PACKET;
			break;

		case ES_VERSION_WRONG:
			break;

		case ES_FW_VERSION_LENGTH:
		case ES_FUNCTION_UNAVAILABLE:
			*error = IV_ERROR_FUNCTION_UNAVAILABLE;
			break;
		}

		return FALSE;
	}

	return res;
}

// Prueft ob Displaydaten in Ordnung.
int CAEChannel::CheckFileDisplay(CDFFileDisplay *pFileDisplay, int *piError)
{
	unsigned long ulFontPtr, ulGraficPtr;
	int i, iFontCount, iGraficCount;
	CDFFileDisplay fileDisplay;

	// Fehlerobjekt loeschen
	*piError = IV_ERROR_SUCCESS;

	// EEPROM - Zeiger auf Zeichensatz lesen 
 	if (!m_channel1_2.ReadMem(MB_START_EEP + 0x0F78, sizeof(unsigned long), (unsigned char*)&ulFontPtr, piError)) {

		return FALSE;
	}
	ChangeOrderLong((unsigned char*)&ulFontPtr);
	if (ulFontPtr == 0xFFFFFFFF) { // Zeiger nicht vorhanden
	
		*piError = IV_ERROR_SUCCESS;
		return FALSE;
	}

	// EEPROM - Zeiger auf Grafik lesen 
 	if (!m_channel1_2.ReadMem(MB_START_EEP + 0x0F7C, sizeof(unsigned long), (unsigned char*)&ulGraficPtr, piError)) {

		return FALSE;
	}
	ChangeOrderLong((unsigned char*)&ulGraficPtr);
	if (ulGraficPtr == 0xFFFFFFFF) { // Zeiger nicht vorhanden
		
		*piError = IV_ERROR_SUCCESS;
		return FALSE;
	}

	// Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE), fileDisplay.m_memFont, piError)) {

		return FALSE;
	}
	// Anzahl Zeichensaetze
	iFontCount = fileDisplay.m_pFontNode->cnt;
	if ((iFontCount > MAX_FONT) || (iFontCount != pFileDisplay->m_pFontNode->cnt)) {

		*piError = IV_ERROR_SUCCESS;
		return FALSE;
	}

	// Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE), fileDisplay.m_memGrafic, piError)) {

		return FALSE;
	}
	// Anzahl Grafiken
	iGraficCount = fileDisplay.m_pGraficNode->cnt;
	if ((iGraficCount > MAX_GRAFIC) || (iGraficCount != pFileDisplay->m_pGraficNode->cnt)) {

		*piError = IV_ERROR_SUCCESS;
		return FALSE;
	}
	// Klassenstrukturen initialisieren
	fileDisplay.Default(iFontCount, iGraficCount);

	// Strukturen aus Flash lesen 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE) + iFontCount * sizeof(unsigned long) + iFontCount * sizeof(FONT), fileDisplay.m_memFont, piError)) {

		return FALSE;
	}
	// CRC von Zeichensaetzen ueberpruefen
	for (i=0; i < iFontCount; i++) {
		if (fileDisplay.m_pFont[i]->crc != pFileDisplay->m_pFont[i]->crc) {

			*piError = IV_ERROR_SUCCESS;
			return FALSE;
		}
	}

	// Strukturen aus Flash lesen 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE) + iGraficCount * sizeof(unsigned long) + iGraficCount * sizeof(GRAFIC), fileDisplay.m_memGrafic, piError)) {

		return FALSE;
	}
	// CRC von Grafiken ueberpruefen
	for (i=0; i < iGraficCount; i++) {
		if (fileDisplay.m_pGrafic[i]->crc != pFileDisplay->m_pGrafic[i]->crc) {

			*piError = IV_ERROR_SUCCESS;
			return FALSE;
		}
	}

	// Zeichensatz- und Grafikdaten vorhanden und identisch
	return TRUE;
}

int CAEChannel::UploadFileDevice(short deviceIDRemote, char *szFileName, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS, 
		protocolVersion = GetProtocolVersion(GetDeviceIDRemote());

	// Puffer fuer die Hardwareinfo in Textform
	memset(m_info, 0, sizeof(m_info));

	//#########################################################################
	// Pruefen ob gewaehlte Datei in das Geraet eingespielt werden kann.
	switch(protocolVersion) {
	case 1: // AEIII+ nicht unterstuetzt
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return FALSE;

	case 2: { // Nur fuer PZE-MasterIV moeglich.

		// Dateityp ermitteln
		struct stat st;
		if (stat(szFileName, &st) == -1) {
			
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}
		
		// Angabe ist eine Datei
		if (S_ISREG(st.st_mode)) {
			
			// Nur ".hex" gueltig.
			int pos = strlen(szFileName) - 4;
#ifdef _WIN32
			if (szFileName[pos] != '.' || 
				((!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				  !(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				  !(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X')) &&
				  (!(szFileName[pos + 1] == 'd' || szFileName[pos + 1] == 'D') ||
				  !(szFileName[pos + 2] == 'f' || szFileName[pos + 2] == 'F') ||
				  !(szFileName[pos + 3] == 'z' || szFileName[pos + 3] == 'Z'))  &&
				  (!(szFileName[pos + 1] == 'z' || szFileName[pos + 1] == 'Z') ||
				  !(szFileName[pos + 2] == 'i' || szFileName[pos + 2] == 'I') ||
				  !(szFileName[pos + 3] == 'p' || szFileName[pos + 3] == 'P'))) ) {
				
				*piError = IV_ERROR_FILE_OPEN;
				return 0;
			}
#else 
			if (!S_ISREG(st.st_mode) || szFileName[pos] != '.' || 
				!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				!(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				!(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X') ) {

				*piError = IV_ERROR_FILE_OPEN;
				return 0;
			}
#endif	
			// Handelt es sich um eine Hexdatei
			if (szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') {

				// Dateidaten laden
				m_fileDevice.Reset();
				if (!m_fileDevice.Load(szFileName) || m_fileDevice.IsCrypt()) {

					m_fileDevice.Reset();

					*piError = IV_ERROR_FILE_OPEN;
					return 0;
				}

				// Ist die Softwareinformation in der Geraetedatei verfuegbar?
				if (m_fileDevice.GetHeaderSoftwareInfo().size) {
					
					// Auslesen der Hardwareinfo
					TDFHardwareInfo hwInfo;
					if (!m_channel1_2.GetHardwareInfo(deviceIDRemote, hwInfo, piError)) {

						return FALSE;
					} // Hardwareinfo
					
					// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
					if (0 == m_fileDevice.CompareInfos(hwInfo, m_fileDevice.GetHeaderSoftwareInfo())) {
						
						// Hardwareinfo in Text aufloesen
						m_channel4.GetInfos(hwInfo, m_info);
						strcpy(m_fileDevice.m_errorExtended, m_fileDevice.m_errorText);
						*piError = IV_ERROR_DEVICEFILE_WRONG;
						return 0;
					}
					
				} else {

					// Die Geraetedatei muss vom Typ "PZE-MasterIV" sein, oder diesen unterstuetzen.
					if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE) {

						// Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
						*piError = IV_ERROR_DEVICEFILE_WRONG;
						return FALSE;
					}
				}


			} else { // Es handelt sich um eine Zipdatei

				// Auslesen der Hardwareinfo
				TDFHardwareInfo hwInfo;
				if (!m_channel1_2.GetHardwareInfo(deviceIDRemote, hwInfo, piError)) {
					
					return FALSE;
				}
				
				// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
				if (0 == m_fileDevice.SearchInfosZip(szFileName, hwInfo)) {
					
					// Hardwareinfo in Text aufloesen
					m_channel4.GetInfos(hwInfo, m_info);
					*piError = IV_ERROR_DEVICEFILE_WRONG;
					return 0;
				}
			}	
		} else if (S_ISDIR(st.st_mode)) {
#ifdef _WIN32			
			// Auswahl aus Verzeichnis nur unter Linux gueltig!
			*piError = IV_ERROR_FUNCTION_PARAMETER;
			return 0;
#else
			// Auslesen der Hardwareinfo
			TDFHardwareInfo hwInfo;
			if (!m_channel1_2.GetHardwareInfo(deviceIDRemote, hwInfo, piError)) {

				return FALSE;
			}
			
			// Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
			if (0 == m_fileDevice.SearchInfosDir(szFileName, hwInfo)) {
				
				// Hardwareinfo in Text aufloesen
				m_channel4.GetInfos(hwInfo, m_info);
				*piError = IV_ERROR_DEVICEFILE_WRONG;
				return 0;
			}
#endif
		} else {
			
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}	
		break;}

	case 3: { // Nur fuer PZE/ZK-MasterIV moeglich.

		// Dateityp ermitteln
		struct stat st;
		if (stat(szFileName, &st) == -1) {

			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}

		// Nur ".hex" gueltig.
		int pos = strlen(szFileName) - 4;
		if (!S_ISREG(st.st_mode) || szFileName[pos] != '.' || 
			!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
			!(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
			!(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X') ) {

			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}

		// Dateidaten laden
		m_fileDevice.Reset();
		if (!m_fileDevice.Load(szFileName) || m_fileDevice.IsCrypt()) {

			m_fileDevice.Reset();

			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}

		unsigned char deviceType = 255;
		// Geraetetyp ermitteln
		if ((iRes = m_channel3.GetDeviceType(deviceIDRemote, deviceType)) != IV_ERROR_SUCCESS) {

			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				iRes = iRes;
			} else if (iRes > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann liegt eine FW 04.01.01.<k vor.
				if (((iRes >> 8) & 0xFF) == 2) {

					// Der Geraetetyp wird manuell gesetzt.
					deviceType = 0;
					iRes = IV_ERROR_SUCCESS;
				} else {

					// Funktionsfehler
					iRes = IV_ERROR_PROCESSING;
				}
			} else {

				// OK
				iRes = IV_ERROR_SUCCESS;
			}

			if (iRes != IV_ERROR_SUCCESS) {

				*piError = iRes;
				return FALSE;
			}
		}

		switch(deviceType) 
		{
		case 0: // PZE-MasterIV
			if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE && 
			!((m_fileDevice.GetHeaderVersionFile() == IV_DFFILEDEVICE_VERSION_MULTIPLE) && 
			 ((1 << deviceType) & m_fileDevice.GetHeaderSoftwareInfo().device)) ) {

				// Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
				*piError = IV_ERROR_DEVICEFILE_WRONG;
				return FALSE;
			}
			break;

		case 1: // ZK - Master IV
			if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_ZK && 
			!((m_fileDevice.GetHeaderVersionFile() == IV_DFFILEDEVICE_VERSION_MULTIPLE) && 
			 ((1 << deviceType) & m_fileDevice.GetHeaderSoftwareInfo().device)) ) {

				// Die gewaehlte Geraetedatei ist NICHT fuer das ZK-MasterIV.
				*piError = IV_ERROR_DEVICEFILE_WRONG;
				return FALSE;
			}
			break;

			// Protokollversion 3 gibt es nur bei PZE-/ZK-MasterIV

		default:
			// Die Setupanwendung ist nicht aktuell genug fuer diesen Geraetetyp.
			*piError = IV_ERROR_DEVICE_UNKNOWN;
			return FALSE;
		}

		break;}

	case 4: {

		// Dateityp ermitteln
		struct stat st;
		if (stat(szFileName, &st) == -1) {
			
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}
		
		// Angabe ist eine Datei
		if (S_ISREG(st.st_mode)) {
			
			// Nur ".hex" gueltig.
			int pos = strlen(szFileName) - 4;
#ifdef _WIN32
			if (szFileName[pos] != '.' || 
				((!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				  !(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				  !(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X')) &&
				  (!(szFileName[pos + 1] == 'd' || szFileName[pos + 1] == 'D') ||
				  !(szFileName[pos + 2] == 'f' || szFileName[pos + 2] == 'F') ||
				  !(szFileName[pos + 3] == 'z' || szFileName[pos + 3] == 'Z')) &&
				  (!(szFileName[pos + 1] == 'z' || szFileName[pos + 1] == 'Z') ||
				  !(szFileName[pos + 2] == 'i' || szFileName[pos + 2] == 'I') ||
				  !(szFileName[pos + 3] == 'p' || szFileName[pos + 3] == 'P'))) ) {
				
				*piError = IV_ERROR_FILE_OPEN;
				return 0;
			}
#else 
			if (!S_ISREG(st.st_mode) || szFileName[pos] != '.' || 
				!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				!(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				!(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X') ) {

				*piError = IV_ERROR_FILE_OPEN;
				return 0;
			}
#endif	
			// Handelt es sich um eine Hexdatei
			if (szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') {
				
				// Dateidaten laden
				m_fileDevice.Reset();
				if (!m_fileDevice.Load(szFileName) || m_fileDevice.IsCrypt()) {
					
					m_fileDevice.Reset();
					
					*piError = IV_ERROR_FILE_OPEN;
					return 0;
				}
				
				// Ist die Softwareinformation in der Geraetedatei verfuegbar?
				if (m_fileDevice.GetHeaderSoftwareInfo().size) {
					
					// Auslesen der Hardwareinfo
					TDFHardwareInfo hwInfo;
					if ((iRes = m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo)) != IV_ERROR_SUCCESS) {
						
						if (iRes < 0) {
							
							// Kommunikationsfehler
							iRes = -iRes;
							// Fehlercode umsetzen
							iRes = iRes;
						} else if (iRes > 0) {
							
							// Wenn der Modulfehler -Funktion unbekannt- ist,
							// dann liegt eine FW 04.01.01.<k vor.
							if (((iRes >> 8) & 0xFF) == 2) {
								
								// Der Geraetetyp wird manuell gesetzt.
								iRes = IV_ERROR_SUCCESS;
							} else {
								
								// Funktionsfehler
								iRes = IV_ERROR_PROCESSING;
							}
						} else {
							
							// OK
							iRes = IV_ERROR_SUCCESS;
						}
						
						if (iRes != IV_ERROR_SUCCESS) {
							
							*piError = iRes;
							return FALSE;
						}
					} // Hardwareinfo
					
					// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
					if (0 == m_fileDevice.CompareInfos(hwInfo, m_fileDevice.GetHeaderSoftwareInfo())) {
						
						// Hardwareinfo in Text aufloesen
						m_channel4.GetInfos(hwInfo, m_info);
						strcpy(m_fileDevice.m_errorExtended, m_fileDevice.m_errorText);
						*piError = IV_ERROR_DEVICEFILE_WRONG;
						return 0;
					}
					
				} else {
					
					unsigned char deviceType = 255;
					// Geraetetyp ermitteln
					if ((iRes = m_channel4.GetTypeDevice(deviceIDRemote, deviceType)) != IV_ERROR_SUCCESS) {
						
						if (iRes < 0) {
							
							// Kommunikationsfehler
							iRes = -iRes;
							// Fehlercode umsetzen
							iRes = iRes;
						} else if (iRes > 0) {
							
							// Wenn der Modulfehler -Funktion unbekannt- ist,
							// dann liegt eine FW 04.01.01.<k vor.
							if (((iRes >> 8) & 0xFF) == 2) {
								
								// Der Geraetetyp wird manuell gesetzt.
								deviceType = 0;
								iRes = IV_ERROR_SUCCESS;
							} else {
								
								// Funktionsfehler
								iRes = IV_ERROR_PROCESSING;
							}
						} else {
							
							// OK
							iRes = IV_ERROR_SUCCESS;
						}
						
						if (iRes != IV_ERROR_SUCCESS) {
							
							*piError = iRes;
							return FALSE;
						}
						
					}
					
					//#######################################################################
					// Wenn die SoftwareInfo fehlt kann die Datei keine Mulitple sein!!!
					
					switch(deviceType) 
					{
					case IV_VALUE_TYPE_DEVICE_PZE: // PZE-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE) {
							
							// Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return FALSE;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_ZK: // ZK-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_ZK) {
							
							// Die gewaehlte Geraetedatei ist NICHT fuer das ZK-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return FALSE;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_BDE: // BDE-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_BDE) {
							
							// Die gewaehlte Geraetedatei ist NICHT fuer das BDE-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return FALSE;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_MOBIL: // Mobil-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MOBIL) {
							
							// Die gewaehlte Geraetedatei ist NICHT fuer das Mobil-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return FALSE;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_MDE: // MDE-BoxIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MDE) {
							
							// Die gewaehlte Geraetedatei ist NICHT fuer das MDE-BoxIV
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return FALSE;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_AE: // AE-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_AE) {
							
							// Die gewaehlte Geraetedatei ist NICHT fuer das AE-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return FALSE;
						}
						break;
						
					default:
						// Die Setupanwendung ist nicht aktuell genug fuer diesen Geraetetyp.
						*piError = IV_ERROR_DEVICE_UNKNOWN;
						return FALSE;
					}
				}
				
			} else { // Es handelt sich um eine Zipdatei

				// Auslesen der Hardwareinfo
				TDFHardwareInfo hwInfo;
				if ((iRes = m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo)) != IV_ERROR_SUCCESS) {
					
					if (iRes < 0) {
						
						// Kommunikationsfehler
						iRes = -iRes;
						// Fehlercode umsetzen
						iRes = iRes;
					} else if (iRes > 0) {
						
						// Wenn der Modulfehler -Funktion unbekannt- ist,
						// dann liegt eine FW 04.01.01.<k vor.
						if (((iRes >> 8) & 0xFF) == 2) {
							
							// Der Geraetetyp wird manuell gesetzt.
							iRes = IV_ERROR_SUCCESS;
						} else {
							
							// Funktionsfehler
							iRes = IV_ERROR_PROCESSING;
						}
					} else {
						
						// OK
						iRes = IV_ERROR_SUCCESS;
					}
					
					if (iRes != IV_ERROR_SUCCESS) {
						
						*piError = iRes;
						return FALSE;
					}
				}
				
				// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
				if (0 == m_fileDevice.SearchInfosZip(szFileName, hwInfo)) {
					
					// Hardwareinfo in Text aufloesen
					m_channel4.GetInfos(hwInfo, m_info);
					*piError = IV_ERROR_DEVICEFILE_WRONG;
					return 0;
				}
			}	
		} else if (S_ISDIR(st.st_mode)) {		
#ifdef _WIN32
			// Auswahl aus Verzeichnis nur unter Linux gueltig!
			*piError = IV_ERROR_FUNCTION_PARAMETER;
			return 0;
#else
			// Auslesen der Hardwareinfo
			TDFHardwareInfo hwInfo;
			if ((iRes = m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo)) != IV_ERROR_SUCCESS) {
				
				if (iRes < 0) {
					
					// Kommunikationsfehler
					iRes = -iRes;
					// Fehlercode umsetzen
					iRes = iRes;
				} else if (iRes > 0) {
					
					// Wenn der Modulfehler -Funktion unbekannt- ist,
					// dann liegt eine FW 04.01.01.<k vor.
					if (((iRes >> 8) & 0xFF) == 2) {
						
						// Der Geraetetyp wird manuell gesetzt.
						iRes = IV_ERROR_SUCCESS;
					} else {
						
						// Funktionsfehler
						iRes = IV_ERROR_PROCESSING;
					}
				} else {
					
					// OK
					iRes = IV_ERROR_SUCCESS;
				}
				
				if (iRes != IV_ERROR_SUCCESS) {
					
					*piError = iRes;
					return FALSE;
				}
			}
			
			// Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
			if (0 == m_fileDevice.SearchInfosDir(szFileName, hwInfo)) {
				
				// Hardwareinfo in Text aufloesen
				m_channel4.GetInfos(hwInfo, m_info);
				*piError = IV_ERROR_DEVICEFILE_WRONG;
				return 0;
			}
#endif
		} else {
			
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}		

		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return FALSE;
	}


	switch(protocolVersion) {

	case 2: { // Protokollversion 2 gibt es nur beim PZE-MasterIV

		// Geraetemodus ermitteln
		int modus = m_channel1_2.GetModus(piError);
		if (!modus) {

			return FALSE;
		}

		// Applikationsmodus? 
		if (modus == 2) {
			// Abfrage ob Daten vorhanden
			if ((iRes = CheckData(deviceIDRemote, piError)) == -1) {

				return FALSE;
			} else if (iRes) {

				*piError = IV_ERROR_DATA_AVAILABLE;
				return FALSE;
			}
		}

		unsigned char loadDisplayData = 1;
		unsigned long sizeData = m_fileDevice.GetSizeData();
		unsigned long tmp = ~0;
		unsigned long address;
		unsigned long firstAddress = m_fileDevice.GetController()->GetProgramFirstAddress();
		unsigned long lastAddress  = m_fileDevice.GetController()->GetProgramLastAddress();
		unsigned long sizeImage   = lastAddress - firstAddress + 1;

		unsigned long crc = 0;

		// EEPROM - Setupzeiger -pKonf- loeschen
		//@@: Uebergangsbug wegen Speicherueberlappung, von Firmware und Setup, beseitigen
 		if (!m_channel1_2.WriteMem(MB_START_EEP + 4, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

			return FALSE;
		}

		// Pruefen ob die Zeichensatz und Grafikdaten vorhanden
		if (!CheckFileDisplay(m_fileDevice.GetFileDisplay(), piError)) {

			if (*piError != IV_ERROR_SUCCESS) {
				// Fehler waehrend Kommunikation, abbruch
				return FALSE;
			}

			// Keine Displaydaten oder ungleich!

		} else {
			// Anzeigedaten vorhanden und Identisch, also nicht uebertragen
			sizeData -= m_fileDevice.GetFileDisplay()->GetSizeData();
			loadDisplayData = 0;
		}

		// EEPROM - Bootload Beginn
 		if (!m_channel1_2.WriteMem(MB_BOOTLOAD, sizeof(unsigned long), (unsigned char*)&sizeData, piError)) {

			return FALSE;
		}

		tmp = FLASH_IMAGE_BASE + firstAddress; // Basisadresse im Flash
		// EEPROM - schreiben: Startadresse des Image im externen Flash
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_SOURCE, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

			return FALSE;
		}

		// EEPROM - schreiben: Zieladresse des Image
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_DEST, sizeof(unsigned long), (unsigned char*)&firstAddress, piError)) {

			return FALSE;
		}

		// EEPROM - schreiben: Laenge des Images
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_LEN, sizeof(unsigned long), (unsigned char*)&sizeImage, piError)) {

			return FALSE;
		}

		// CRC berechnen
		crc = GetCRC16(m_fileDevice.GetController()->GetMemProgram(firstAddress), sizeImage);

		// EEPROM - schreiben: CRC des Images
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_CRC, sizeof(unsigned long), (unsigned char*)&crc, piError)) {

			return FALSE;
		}

		// Imagedaten
		for (address=firstAddress; address < (lastAddress + 1); address += MAX_MB_PAKET_DATA) {

			// FLASH - schreiben
 			if (!m_channel1_2.WriteMem(FLASH_IMAGE_BASE + address, 
				(((lastAddress + 1) - address) > MAX_MB_PAKET_DATA) ? MAX_MB_PAKET_DATA : ((lastAddress + 1) - address), 
				m_fileDevice.GetController()->GetMemProgram(address), piError)) {

				return FALSE;
			}

		} 

		// Laden der Zeichensatz- und Grafikdaten?
		if (loadDisplayData) {

			// Zeichensatz- und Grafikdaten
			CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();

			// Zeichensatzdaten schreiben
			for (address=0; address < pFileDisplay->m_sizeFont; address += MAX_MB_PAKET_DATA) {

				// Flash schreiben 
				if (!m_channel1_2.WriteMem(0x000B0200 + address, 
					((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeFont) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeFont - address, 
					&pFileDisplay->m_memFont[address], piError)) {

					return FALSE;
				}
			}

			// Grafikdaten schreiben
			for (address=0; address < pFileDisplay->m_sizeGrafic; address += MAX_MB_PAKET_DATA) {

				if (!m_channel1_2.WriteMem(0x000B0200 + pFileDisplay->m_sizeFont + address, 
					((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeGrafic) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeGrafic - address,
					&pFileDisplay->m_memGrafic[address], piError)) {

					return FALSE;
				}
			}

			tmp = 0;
			ChangeOrderLong((unsigned char*)&tmp);
			// EEPROM - Zeiger auf Font setzen 
 			if (!m_channel1_2.WriteMem(MB_START_EEP + 0x0F78, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

				return FALSE;
			}

			tmp = pFileDisplay->m_sizeFont;
			ChangeOrderLong((unsigned char*)&tmp);
			// EEPROM - Zeiger auf Grafik setzen 
 			if (!m_channel1_2.WriteMem(MB_START_EEP + 0x0F7C, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

				return FALSE;
			}
		}

		// Schreiben der Geraetetexte
		// Noch einbauen!! 

		sizeData = 0;
		// EEPROM - Bootload Ende
 		if (!m_channel1_2.WriteMem(MB_BOOTLOAD, sizeof(unsigned long), (unsigned char*)&sizeData, piError)) {

			return FALSE;
		}

		// Protokollversion ist aufgrund eines Kommunikationsfehlers auf 0 zu setzen.
		// Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
		m_protocolVersion[(GetDeviceIDRemote() == 0xFE) ? 0 : GetDeviceIDRemote()] = 0;

		return TRUE;}

	case 3: { // PZE-/ZK-MasterIV mit Protokollversion 3

		char version[64];
		int modus = 0;
		// Geraetemodus (Application/Bootloader) ermitteln
		if ((iRes = m_channel3.GetVersion(deviceIDRemote, version)) != IV_ERROR_SUCCESS) {

			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				iRes = iRes;
			} else if (iRes > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann befindet sich das Geraet im Bootloadermodus.
				if (((iRes >> 8) & 0xFF) == 2) {

					// Bootloadermodus
					modus = 1;
					iRes = IV_ERROR_SUCCESS;
				} else {

					// Funktionsfehler
					iRes = IV_ERROR_PROCESSING;
				}
			}
		}

		iRes = 0;
		// Abfrage nur wenn Geraet im Applicationmodus
		// Abfrage ob Daten vorhanden
		if (!modus && (iRes = CheckData(deviceIDRemote, piError)) == -1) {

			return FALSE;
		} else if (iRes) {

			*piError = IV_ERROR_DATA_AVAILABLE;
			return FALSE;
		}

		iRes = UploadDevice3(deviceIDRemote);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Funktionsfehler
			*piError = IV_ERROR_PROCESSING;
			iRes = 0;
		} else {

			// OK
			iRes = 1;

			// Protokollversion ist aufgrund eines Kommunikationsfehlers auf 0 zu setzen.
			// Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
			m_protocolVersion[(GetDeviceIDRemote() == 0xFE) ? 0 : GetDeviceIDRemote()] = 0;
		}
		return iRes;}

	case 4: { // MasterIV mit Protokollversion 4

		//######################################################################
		// 1. Ermittlung ob auf dem Geraet eine Applicationfirmware laeuft.
		// 2. Wenn Punkt 1 == TRUE, dann wird geprueft ob Datensaetze verfuegbar sind.
		// 2. Wenn Punkt 2 == TRUE, wird die Aktion abgebrochen.
		char version[64];
		int modus = 0;
		// Geraetemodus (Application/Bootloader) ermitteln
		if ((iRes = m_channel4.GetVersion(deviceIDRemote, version)) != IV_ERROR_SUCCESS) {

			if (iRes < 0) {

				// Kommunikationsfehler
				iRes = -iRes;
				// Fehlercode umsetzen
				iRes = iRes;
			} else if (iRes > 0) {

				// Wenn der Modulfehler -Funktion unbekannt- ist,
				// dann befindet sich das Geraet im Bootloadermodus.
				if (((iRes >> 8) & 0xFF) == 2) {

					// Bootloadermodus
					modus = 1;
					iRes = IV_ERROR_SUCCESS;
				} else {

					// Funktionsfehler
					iRes = IV_ERROR_PROCESSING;
				}
			}
		}

		iRes = 0;
		// Abfrage nur wenn Geraet im Applicationmodus
		// Abfrage ob Daten vorhanden
		if (!modus && (iRes = CheckData(deviceIDRemote, piError)) == -1) {

			return FALSE;
		} else if (iRes) {

			*piError = IV_ERROR_DATA_AVAILABLE;
			return FALSE;
		}

		iRes = UploadDevice4(deviceIDRemote);
		if (iRes < 0) {

			// Kommunikationsfehler
			iRes = -iRes;
			// Fehlercode umsetzen
			*piError = iRes;
			iRes = 0;
		} else if (iRes > 0) {

			// Funktionsfehler
			*piError = IV_ERROR_PROCESSING;
			iRes = 0;
		} else {

			// OK
			iRes = 1;

			// Protokollversion ist aufgrund eines Kommunikationsfehlers auf 0 zu setzen.
			// Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
			m_protocolVersion[(GetDeviceIDRemote() == 0xFE) ? 0 : GetDeviceIDRemote()] = 0;
		}
		return iRes;}
	}

	return FALSE;
}

int CAEChannel::UploadDevice3(short deviceIDRemote)
{
	// Zeichensatz- und Grafikdaten
	CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();
	int res = IV_ERROR_SUCCESS;
	unsigned long address = 0;
	unsigned char *buf;
	unsigned long size = 0;
	unsigned long crc = 0;
	char version[64];

	memset(version, 0, sizeof(version));
	// Bootversion lesen 
	if ((res = m_channel3.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Versionsstring: xx.xx.xx.xx
	// Protokollversion 2:
		// Das Geraet kann sich hier nicht im Bootloadermodus befinden,
		// da die Befehle auf Protokollversion 2 basieren.
	if ((strstr(version, "00.02.00") != version) &&
	// Protokollversion 3:
		// Das Geraet kann sich im Bootloadermodus befinden.
		(strstr(version, "00.03.00") != version) &&
	// Protokollversion 4:
		// Das Geraet kann sich im Bootloadermodus befinden.
		(strstr(version, "00.04.00") != version) ) {

		return -IV_ERROR_BOOTVERSION;
	}

	// Systemstopp setzen
	if ((res = m_channel3.BeginSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}


	// Sind Displaydaten vorhanden?
	if (pFileDisplay->GetSizeData()) {

		// ZEICHENSAETZE

		address = 0;
		buf = pFileDisplay->m_memFont;
		size = pFileDisplay->m_sizeFont;
		crc = m_channel3.GetCRC32(buf, size);

		// Systemstopp setzen
		if ((res = m_channel3.BeginSegment(deviceIDRemote, IV_VALUE_SEGMENT_FONTS, address, size, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Zeichensatzdaten uebertragen
		if ((res = m_channel3.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Systemstopp aufheben
		if ((res = m_channel3.EndSegment(deviceIDRemote, IV_VALUE_SEGMENT_FONTS, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}


		// GRAFIKEN

		address = 0;
		buf = pFileDisplay->m_memGrafic;
		size = pFileDisplay->m_sizeGrafic;
		crc = m_channel3.GetCRC32(buf, size);

		// Systemstopp setzen
		if ((res = m_channel3.BeginSegment(deviceIDRemote, IV_VALUE_SEGMENT_GRAFICS, address, size, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Grafikdaten uebertragen
		if ((res = m_channel3.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Systemstopp aufheben
		if ((res = m_channel3.EndSegment(deviceIDRemote, IV_VALUE_SEGMENT_GRAFICS, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}
	}

	// FIRMWARE

	address = 0;
	buf = m_fileDevice.GetController()->GetMemProgram(m_fileDevice.GetController()->GetProgramFirstAddress());
	size   = m_fileDevice.GetController()->GetProgramLastAddress() - m_fileDevice.GetController()->GetProgramFirstAddress() + 1;
	crc = m_channel3.GetCRC32(buf, size);

	// Systemstopp setzen
	if ((res = m_channel3.BeginSegment(deviceIDRemote, IV_VALUE_SEGMENT_FIRMWARE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Firmwaredaten uebertragen
	if ((res = m_channel3.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Systemstopp aufheben
	if ((res = m_channel3.EndSegment(deviceIDRemote, IV_VALUE_SEGMENT_FIRMWARE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Schreiben der Geraetetexte
	/* Noch einbauen!! */

	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel3.EndSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	return res;
}

int CAEChannel::UploadDevice4(short deviceIDRemote)
{
	// Zeichensatz- und Grafikdaten
	CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();
	int res = IV_ERROR_SUCCESS;
	unsigned long address = 0;
	unsigned char *buf;
	unsigned long size = 0;
	unsigned long crc = 0;
	char version[64];

	memset(version, 0, sizeof(version));
	// Bootversion lesen 
	if ((res = m_channel4.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}


	// Sind Displaydaten vorhanden?
	if (pFileDisplay->GetSizeData()) {

		// ZEICHENSAETZE

		address = 0;
		buf = pFileDisplay->m_memFont;
		size = pFileDisplay->m_sizeFont;
		crc = m_channel4.GetCRC32(buf, size);

		// Systemstopp setzen
		if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FONTS, address, size, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Zeichensatzdaten uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Systemstopp aufheben
		if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FONTS, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}


		// GRAFIKEN

		address = 0;
		buf = pFileDisplay->m_memGrafic;
		size = pFileDisplay->m_sizeGrafic;
		crc = m_channel4.GetCRC32(buf, size);

		// Systemstopp setzen
		if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GRAFICS, address, size, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Grafikdaten uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

			return res;
		}

		// Systemstopp aufheben
		if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GRAFICS, crc)) != IV_ERROR_SUCCESS) {

			return res;
		}
	}

	// FIRMWARE

	address = 0;
	buf = m_fileDevice.GetController()->GetMemProgram(m_fileDevice.GetController()->GetProgramFirstAddress());
	size   = m_fileDevice.GetController()->GetProgramLastAddress() - m_fileDevice.GetController()->GetProgramFirstAddress() + 1;
	crc = m_channel4.GetCRC32(buf, size);

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FIRMWARE, address, size, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Firmwaredaten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FIRMWARE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	// Schreiben der Geraetetexte
	/* Noch einbauen!! */

	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}

	return res;
}


//#########################################################
//######## Schnittstelle zu Channel1/Channel2 #############
//#########################################################

int CAEChannel::GetModus(short deviceIDRemote, int *error)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 3:
	case 4: return -IV_ERROR_VERSION_FIRMWARE;
	case 2: return m_channel1_2.GetModus(error);
	}

	return -IV_ERROR_UNREACHABLE;
}


//#########################################################
//######## Schnittstelle zu Channel3/Channel4 #############
//#########################################################

void CAEChannel::SetCBFunctionWithParams(int (*pFunction)(unsigned long, unsigned long))
{
	m_channel3.SetCBFunctionWithParams(pFunction);
	m_channel4.SetCBFunctionWithParams(pFunction);
}

unsigned long CAEChannel::GetCRC32(unsigned char *data, unsigned long size, unsigned long crc, unsigned char closure)
{
	return m_channel4.GetCRC32(data, size, crc, closure);
}

// MODULE_HARDWARE

// Einfache Anfrage ob das Geraet antwortet.
int CAEChannel::Ping(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.Ping(deviceIDRemote);
	case 4: return m_channel4.Ping(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeDevice(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetDeviceType(deviceIDRemote, type);
	case 4: return m_channel4.GetTypeDevice(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeBoard(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetBoardType(deviceIDRemote, type);
	case 4: return m_channel4.GetTypeBoard(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentRS485_0(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetComponentRS485_0(deviceIDRemote, value);
	case 4: return m_channel4.GetComponentRS485_0(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentTCPIP(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetComponentTCPIP(deviceIDRemote, value);
	case 4: return m_channel4.GetComponentTCPIP(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeModemMC(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetComponentModemMC35i(deviceIDRemote, value);
	case 4: return m_channel4.GetTypeModemMC(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentRS485_1(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetComponentRS485_1(deviceIDRemote, value);
	case 4: return m_channel4.GetComponentRS485_1(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeDisplay(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeDisplay(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeKeypad(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeKeypad(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeTransponder(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeTransponder(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeMMC(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeMMC(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeRadio(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeRadio(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeGPS(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeGPS(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentUSB(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetComponentUSB(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeDCF77(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeDCF77(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentInAnalog(short deviceIDRemote, unsigned long &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetComponentInAnalog(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentInDigital(short deviceIDRemote, unsigned long &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetComponentInDigital(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentOutDigital(short deviceIDRemote, unsigned long &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetComponentOutDigital(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeFingerprint(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeFingerprint(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeExternalReader(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetTypeExternalReader(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionIOModule(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetOptionIOModule(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionTBAdaption(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetOptionTBAdaption(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetHWConfig(short deviceIDRemote, unsigned char &value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetHWConfig(deviceIDRemote, value);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetHardwareInfo(short deviceIDRemote, TDFHardwareInfo &hwInfo)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetInfos(TDFHardwareInfo hwInfo, char *info)
{
	return m_channel4.GetInfos(hwInfo, info);
}

// MODULE_SYSTEM

// Systemreset; Speicherbereiche loeschen
//int SystemReset(short deviceIDRemote, unsigned short modus);

// Firmwareversion lesen
int CAEChannel::GetVersion(short deviceIDRemote, char *version)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetVersion(deviceIDRemote, version);
	case 4: return m_channel4.GetVersion(deviceIDRemote, version);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetVersionBootloader(short deviceIDRemote, char *version)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1: return -IV_ERROR_VERSION_FIRMWARE;
	case 2: {
		unsigned long ver;

		// Fehlerobjekt loeschen
		int error = IV_ERROR_SUCCESS;

		// EEPROM - Bootversion lesen 
 		if (!m_channel1_2.ReadMem(MB_START_EEP + EEPROM_BOOT_VERSION, sizeof(unsigned long), (unsigned char*)&ver, &error)) {

			return FALSE;
		}

		sprintf(version, "%02X.%02X.%02X.%02X", ((unsigned char*)&ver)[3], ((unsigned char*)&ver)[2], ((unsigned char*)&ver)[1], ((unsigned char*)&ver)[0]);

		return IV_ERROR_SUCCESS;}
	case 3: return m_channel3.GetVersionBootloader(deviceIDRemote, version);
	case 4: return m_channel4.GetVersionBootloader(deviceIDRemote, version);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetSerial(short deviceIDRemote, unsigned long &serial)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetSerial(deviceIDRemote, serial);
	case 4: return m_channel4.GetSerial(deviceIDRemote, serial);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::OpenComServerMode(short deviceIDRemote, char *comm, unsigned long baudRate, unsigned char byteSize, unsigned char parity, unsigned char stopBits, unsigned short flags, unsigned short timeoutToClose)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2:return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.OpenComServerMode(deviceIDRemote, comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::CloseComServerMode(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	return m_channel4.CloseComServerMode(deviceIDRemote);
}


// Datum/Uhrzeit setzen/lesen
//int SetRTC(short deviceIDRemote, char *time);
//int GetRTC(short deviceIDRemote, char *time);

// Systemvariable setzen/lesen
//int SetSystemVariable(short deviceIDRemote, char *name, char *value);
//int GetSystemVariable(short deviceIDRemote, char *name, char *value);

// Relais setzen/lesen
//int SetRelay(short deviceIDRemote, char *name, unsigned char busNum, unsigned char relayNum, unsigned char status, unsigned short time);
//int GetRelay(short deviceIDRemote, char *name, unsigned char busNum, unsigned char relayNum, unsigned char &status, unsigned short &time);

// Datensegmentuebertragung beginnen/enden
int CAEChannel::BeginSegment(short deviceIDRemote, unsigned short art, unsigned long &address, unsigned long size, unsigned long crc)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.BeginSegment(deviceIDRemote, art, address, size, crc);
	case 4: return m_channel4.BeginSetSegment(deviceIDRemote, art, address, size, crc);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::EndSegment(short deviceIDRemote, unsigned short art, unsigned long crc)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.EndSegment(deviceIDRemote, art, crc);
	case 4: return m_channel4.EndSetSegment(deviceIDRemote, art, crc);
	}

	return -IV_ERROR_UNREACHABLE;
}


// Uebertragen von Speicherdaten in Datei.
int CAEChannel::TransferMemToFile(short deviceIDRemote, int type, long address, int size, char *fileName)
{
	StructMemFileHeader header;
	int res = IV_ERROR_SUCCESS, chunk;
	char buf[0x400];
	long baseAddress = 0, offsAddress = 0;

	FILE *fp = fopen(fileName, "wb");
	if (fp == NULL) {
		return -IV_ERROR_FILE_OPEN;
	}

	// Remoteadresse einstellen
	SetDeviceIDRemote(deviceIDRemote);

	// Dateiheader aufsetzen
	memset(&header, 0, sizeof(header));
	header.structSize = sizeof(header);
	header.type = type;
	header.baseAddress = baseAddress;
	header.memAddress = address;
	header.memSize = size;

	// Aufgrund Protokollversion entscheiden
	// Protokollverwendung pruefen
	switch (GetProtocolVersion(m_channel4.GetDeviceIDRemote())) {
	case 1:
	case 2:
		switch(type) {
		case 0:
			baseAddress = MB_ACCESS_SRAM;
			break;
		case 1:
			baseAddress = MB_ACCESS_EEPROM;
			break;
		case 2:
			baseAddress = MB_ACCESS_FLASH;
			break;
		case 3:
			baseAddress = MB_ACCESS_PROGRAM;
			break;
		}

		// Basisaddress nachtragen
		header.baseAddress = baseAddress;
		// Dateiheader schreiben
		if (fwrite(&header, sizeof(header), 1, fp) != 1) {
		
			res = -IV_ERROR_FAIL;
			goto label_out;
		}

		for ( ;; ) {
			// Naechsten Block berechnen
			chunk = ((size - offsAddress) < sizeof(buf)) ? chunk = size - offsAddress : chunk = sizeof(buf);
			if (chunk == 0) break; // Keine weiteren Daten

			if (!m_channel1_2.ReadMem(baseAddress + address + offsAddress, chunk, (unsigned char*)buf, &res)) {

				res = -IV_ERROR_FAIL;
				goto label_out;
			}

			if (fwrite(buf, chunk, 1, fp) != 1) {

				res = -IV_ERROR_FAIL;
				goto label_out;
			}

			offsAddress += chunk;
		}
		break;

	case 4:
		// Dateiheader schreiben
		if (fwrite(&header, sizeof(header), 1, fp) != sizeof(header)) {
		
			return -IV_ERROR_FAIL;
		}

		switch(type) {
		case 0:
			break;

		case 1:
			if ((res = m_channel4.GetEEPROM(deviceIDRemote, address, (unsigned char*)buf, size)) != IV_ERROR_SUCCESS) {

				goto label_out;
			}
			break;
		case 2:
			if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)buf, size)) != IV_ERROR_SUCCESS) {

				goto label_out;
			}
			break;

		case 3:
			break;
		}
		break;

	default:
		res = -IV_ERROR_UNREACHABLE;
		goto label_out;
	}

label_out:
	if (fp) fclose(fp);
	return res;
}

// Uebertragen von Speicherdaten aus einer Datei.
int CAEChannel::TransferFileToMem(short deviceIDRemote, char *fileName)
{
	StructMemFileHeader header;
	int res = IV_ERROR_SUCCESS, chunk = 0;
	long baseAddress = 0, offsAddress = 0;
	char buf[0x400];
	
	FILE *fp = fopen(fileName, "rb");
	if (fp == NULL) {
		return -IV_ERROR_FILE_OPEN;
	}

	// Remoteadresse einstellen
	SetDeviceIDRemote(deviceIDRemote);

	// Dateiheader lesen
	if (fread(&header, sizeof(header), 1, fp) != sizeof(header)) {
	
		res = -IV_ERROR_FAIL;
		goto label_out;
	}

	// Aufgrund Protokollversion entscheiden
	// Protokollverwendung pruefen
	switch (GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: {
		offsAddress = header.baseAddress;
		for ( ;; ) {
			// Naechsten Block berechnen
			chunk = ((header.memSize - offsAddress) < sizeof(buf)) ? chunk = header.memSize - offsAddress : chunk = sizeof(buf);

			if (fread(buf, chunk, 1, fp) != (unsigned)chunk) {

				res = -IV_ERROR_FAIL;
				goto label_out;
			}

			if (!m_channel1_2.WriteMem(header.baseAddress + header.memAddress + offsAddress, chunk, (unsigned char*)buf, &res)) {

				goto label_out;
			}


			offsAddress += chunk;
		}
		break;}

	case 4:
		switch(header.type) {
		case 0://MB_ACCESS_SRAM:
			break;

		case 1://MB_ACCESS_EEPROM:
			if ((res = m_channel4.SetEEPROM(deviceIDRemote, header.memAddress, (unsigned char*)buf, chunk)) != IV_ERROR_SUCCESS) {

				goto label_out;
			}
			break;

		case 2://MB_ACCESS_FLASH:
			if ((res = m_channel4.SetFlash(deviceIDRemote, header.memAddress, (unsigned char*)buf, chunk)) != IV_ERROR_SUCCESS) {

				goto label_out;
			}
			break;

		case 3://MB_ACCESS_PROGRAM:
			break;
		}
		break;
	default:
		res = -IV_ERROR_UNREACHABLE;
		goto label_out;
	}

label_out:
	if (fp) fclose(fp);
	return res;
}

// Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
int CAEChannel::SetFlash(short deviceIDRemote, unsigned long address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int err;
	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1: return m_channel1_2.WriteMem(address, count, buf, &err, 32);
	case 2: return m_channel1_2.WriteMem(address, count, buf, &err);
	case 3: return m_channel3.SetFlash(deviceIDRemote, address, buf, count);
	case 4: return m_channel4.SetFlash(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetFlash(short deviceIDRemote, unsigned long address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int err;
	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1: return m_channel1_2.ReadMem(address, count, buf, &err);
	case 2: return m_channel1_2.ReadMem(address, count, buf, &err);
	case 3: return m_channel3.GetFlash(deviceIDRemote, address, buf, count);
	case 4: return m_channel4.GetFlash(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetFlashPage(short deviceIDRemote, unsigned short pageNum, unsigned char *buf, int &count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1: 
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetFlashPage(deviceIDRemote, pageNum, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

// EEPROM setzen/lesen 
int CAEChannel::SetEEPROM(short deviceIDRemote, unsigned long address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.SetEEPROM(deviceIDRemote, address, buf, count);
	case 4: return m_channel4.SetEEPROM(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetEEPROM(short deviceIDRemote, unsigned long address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetEEPROM(deviceIDRemote, address, buf, count);
	case 4: return m_channel4.GetEEPROM(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

// PROGRAM setzen/(lesen unterbunden)
//int SetProgram(short deviceIDRemote, unsigned long address, unsigned char *buf, unsigned short count);
//int GetProgram(short deviceIDRemote, unsigned long address, unsigned char *buf, unsigned short count);

// Systembios aufrufen/verlassen
int CAEChannel::EnterBios(short deviceIDRemote, unsigned char protocolVersion)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(protocolVersion) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.EnterBios(deviceIDRemote);
	case 4: return m_channel4.EnterBios(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::ExitBios(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.ExitBios(deviceIDRemote);
	case 4: return m_channel4.ExitBios(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Interface setzen/lesen (RS232, TCPIP...)
int CAEChannel::SetInterface(short deviceIDRemote, unsigned char type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.SetInterface(deviceIDRemote, type);
	case 4: return m_channel4.SetInterface(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetInterface(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetInterface(deviceIDRemote, type);
	case 4: return m_channel4.GetInterface(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int CAEChannel::SetBaud(short deviceIDRemote, unsigned long br)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.SetBaud(deviceIDRemote, br);
	case 4: return m_channel4.SetBaud(deviceIDRemote, br);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetBaud(short deviceIDRemote, unsigned long &br)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetBaud(deviceIDRemote, br);
	case 4: return m_channel4.GetBaud(deviceIDRemote, br);
	}

	return -IV_ERROR_UNREACHABLE;
}

// DeviceID setzen/lesen
int CAEChannel::SetDeviceIDBios(short deviceIDRemote, unsigned short id)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.SetDeviceIDBios(deviceIDRemote, id);
	case 4: return m_channel4.SetDeviceIDBios(deviceIDRemote, id);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetDeviceIDBios(short deviceIDRemote, unsigned short &id)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.GetDeviceIDBios(deviceIDRemote, id);
	case 4: return m_channel4.GetDeviceIDBios(deviceIDRemote, id);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Bootloader aktivieren
int CAEChannel::EnterBootloader(short deviceIDRemote, unsigned char protocolVersion)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(protocolVersion) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.EnterBootloader(deviceIDRemote);
	case 4: return m_channel4.EnterBootloader(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Monitor Mode vom XPort
int CAEChannel::EnterXPortMonitor(short deviceIDRemote, char *data)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.EnterXPortMonitor(deviceIDRemote, data);
	case 4: return m_channel4.EnterXPortMonitor(deviceIDRemote, data);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::CommandXPort(short deviceIDRemote, char *buf, int &count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.CommandXPort(deviceIDRemote, buf, (unsigned short&)count);
	case 4: return m_channel4.CommandXPort(deviceIDRemote, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::ExitXPortMonitor(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return m_channel3.ExitXPortMonitor(deviceIDRemote);
	case 4: return m_channel4.ExitXPortMonitor(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Softwareinformation der Anwendungsfirmware abfragen
int CAEChannel::GetSoftwareInfo(short deviceIDRemote, unsigned char *swi, int &infoLength)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetSoftwareInfo(deviceIDRemote, swi, infoLength);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Oeffnen einer Tabelle
int CAEChannel::TableOpen(short deviceIDRemote, char *name, int options, int &handle, char *reserved)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableOpen(deviceIDRemote, name, options, handle, reserved);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Schliessen der Tabelle
int CAEChannel::TableClose(short deviceIDRemote, int handle)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableClose(deviceIDRemote, handle);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Setzen eines Filters auf die Tabelle
int CAEChannel::TableSetFilter(short deviceIDRemote, int handle, char *filter)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableSetFilter(deviceIDRemote, handle, filter);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Lesen des Filters auf die Tabelle
int CAEChannel::TableGetFilter(short deviceIDRemote, int handle, char *filter, short &len)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableGetFilter(deviceIDRemote, handle, filter, len);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Loeschen des Filters der Tabelle
int CAEChannel::TableClearFilter(short deviceIDRemote, int handle)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableClearFilter(deviceIDRemote, handle);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
int CAEChannel::TableSkip(short deviceIDRemote, int handle, int offset, short origin)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableSkip(deviceIDRemote, handle, offset, origin);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int CAEChannel::TableSetField(short deviceIDRemote, int handle, char *name, char *value)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableSetField(deviceIDRemote, handle, name, value);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int CAEChannel::TableGetField(short deviceIDRemote, int handle, char *name, char *value, short &len)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.TableGetField(deviceIDRemote, handle, name, value, len);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int CAEChannel::GetInfo(short deviceIDRemote, char *name, char *params, char *info, int &len)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.GetInfo(deviceIDRemote, name, params, info, len);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Zutrittsbuchung durchfuehren
int CAEChannel::Entrance2Identification(short deviceIDRemote, char *TM, char *Id, char *Pin, char *status, int &statusLength)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.Entrance2Identification(deviceIDRemote, TM, Id, Pin, status, statusLength);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetNotifyOnDevice(short deviceIDRemote, int type, int state)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.SetNotify(deviceIDRemote, type, state);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Datensatz an Fingerprint-Tabelle anfuegen.
int CAEChannel::FingerprintAppendRecord(short deviceIDRemote, unsigned char type, unsigned char *fingertemplate, int size)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.FingerprintTableAppendRecord(deviceIDRemote, type, fingertemplate, size);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle lesen.
int CAEChannel::FingerprintGetRecord(short deviceIDRemote, unsigned char &type, int pid, int fid, unsigned char *fingertemplate, int &size)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.FingerprintTableGetRecord(deviceIDRemote, type, pid, fid, fingertemplate, size);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle loeschen.
int CAEChannel::FingerprintDeleteRecord(short deviceIDRemote, int pid, int fid)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(deviceIDRemote)) {
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 3: return (2 << 8); // Modulfehler: Die geforderte Funktion wird nicht unterstuetzt.
	case 4: return m_channel4.FingerprintTableDeleteRecord(deviceIDRemote, pid, fid);
	}

	return -IV_ERROR_UNREACHABLE;
}


#ifndef DATAFOX_STUDIO_IV

// Funktions-Stumps

// MODULE_HARDWARE

int CAEChannel::SetTypeDevice(unsigned char deviceIDRemote, unsigned char type)           { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeBoard(unsigned char deviceIDRemote, unsigned char type)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeModemMC(unsigned char deviceIDRemote, unsigned char value)         { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeDisplay(unsigned char deviceIDRemote, unsigned char type)          { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeKeypad(unsigned char deviceIDRemote, unsigned char type)           { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeTransponder(unsigned char deviceIDRemote, unsigned char type)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeMMC(unsigned char deviceIDRemote, unsigned char type)              { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeRadio(unsigned char deviceIDRemote, unsigned char type)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeGPS(unsigned char deviceIDRemote, unsigned char type)              { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeDCF77(unsigned char deviceIDRemote, unsigned char type)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeFingerprint(unsigned char deviceIDRemote, unsigned char type)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeExternalReader(unsigned char deviceIDRemote, unsigned char value)  { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

int CAEChannel::SetComponentRS485_0(unsigned char deviceIDRemote, unsigned char value)    { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentTCPIP(unsigned char deviceIDRemote, unsigned char value)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentRS485_1(unsigned char deviceIDRemote, unsigned char value)    { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentUSB(unsigned char deviceIDRemote, unsigned char value)        { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentInAnalog(unsigned char deviceIDRemote, unsigned long value)   { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentInDigital(unsigned char deviceIDRemote, unsigned long value)  { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentOutDigital(unsigned char deviceIDRemote, unsigned long value) { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

int CAEChannel::SetOptionIOModule(unsigned char deviceIDRemote, unsigned char value)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetOptionTBAdaption(unsigned char deviceIDRemote, unsigned char value)    { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

int CAEChannel::SetHWConfig(unsigned char deviceIDRemote, unsigned char value)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }


// MODULE_SYSTEM

int CAEChannel::SetSerial(unsigned char deviceIDRemote, unsigned long serial)             { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

#endif
